from collections.abc import Callable
from io import StringIO
from logging import Logger
from time import sleep
from typing import cast

from openai import OpenAI
from openai.types.chat import ChatCompletionMessageParam

from .error import is_retry_error
from .increasable import Increasable, Increaser
from .types import Message, MessageRole, R


class LLMExecutor:
    def __init__(
        self,
        api_key: str,
        url: str,
        model: str,
        timeout: float | None,
        top_p: Increasable,
        temperature: Increasable,
        retry_times: int,
        retry_interval_seconds: float,
        create_logger: Callable[[], Logger | None],
    ) -> None:
        self._model_name: str = model
        self._timeout: float | None = timeout
        self._top_p: Increasable = top_p
        self._temperature: Increasable = temperature
        self._retry_times: int = retry_times
        self._retry_interval_seconds: float = retry_interval_seconds
        self._create_logger: Callable[[], Logger | None] = create_logger
        self._client = OpenAI(
            api_key=api_key,
            base_url=url,
            timeout=timeout,
        )

    def request(self, messages: list[Message], parser: Callable[[str], R], max_tokens: int | None) -> R:
        result: R | None = None
        last_error: Exception | None = None
        did_success = False
        top_p: Increaser = self._top_p.context()
        temperature: Increaser = self._temperature.context()
        logger = self._create_logger()

        if logger is not None:
            logger.debug(f"[[Request]]:\n{self._input2str(messages)}\n")

        try:
            for i in range(self._retry_times + 1):
                try:
                    response = self._invoke_model(
                        input_messages=messages,
                        top_p=top_p.current,
                        temperature=temperature.current,
                        max_tokens=max_tokens,
                    )
                    if logger is not None:
                        logger.debug(f"[[Response]]:\n{response}\n")

                except Exception as err:
                    last_error = err
                    if not is_retry_error(err):
                        raise err
                    if logger is not None:
                        logger.warning(f"request failed with connection error, retrying... ({i + 1} times)")
                    if self._retry_interval_seconds > 0.0 and i < self._retry_times:
                        sleep(self._retry_interval_seconds)
                    continue

                try:
                    result = parser(response)
                    did_success = True
                    break

                except Exception as err:
                    last_error = err
                    warn_message = f"request failed with parsing error, retrying... ({i + 1} times)"
                    if logger is not None:
                        logger.warning(warn_message)
                    print(warn_message)
                    top_p.increase()
                    temperature.increase()
                    if self._retry_interval_seconds > 0.0 and i < self._retry_times:
                        sleep(self._retry_interval_seconds)
                    continue

        except KeyboardInterrupt as err:
            if last_error is not None and logger is not None:
                logger.debug(f"[[Error]]:\n{last_error}\n")
            raise err

        if not did_success:
            if last_error is None:
                raise RuntimeError("Request failed with unknown error")
            else:
                raise last_error

        return cast(R, result)

    def _input2str(self, input: str | list[Message]) -> str:
        if isinstance(input, str):
            return input
        if not isinstance(input, list):
            raise ValueError(f"Unsupported input type: {type(input)}")

        buffer = StringIO()
        is_first = True
        for message in input:
            if not is_first:
                buffer.write("\n\n")
            if message.role == MessageRole.SYSTEM:
                buffer.write("System:\n")
                buffer.write(message.message)
            elif message.role == MessageRole.USER:
                buffer.write("User:\n")
                buffer.write(message.message)
            elif message.role == MessageRole.ASSISTANT:
                buffer.write("Assistant:\n")
                buffer.write(message.message)
            else:
                buffer.write(str(message))
            is_first = False

        return buffer.getvalue()

    def _invoke_model(
        self,
        input_messages: list[Message],
        top_p: float | None,
        temperature: float | None,
        max_tokens: int | None,
    ):
        messages: list[ChatCompletionMessageParam] = []
        for item in input_messages:
            if item.role == MessageRole.SYSTEM:
                messages.append(
                    {
                        "role": "system",
                        "content": item.message,
                    }
                )
            elif item.role == MessageRole.USER:
                messages.append(
                    {
                        "role": "user",
                        "content": item.message,
                    }
                )
            elif item.role == MessageRole.ASSISTANT:
                messages.append(
                    {
                        "role": "assistant",
                        "content": item.message,
                    }
                )

        stream = self._client.chat.completions.create(
            model=self._model_name,
            messages=messages,
            stream=True,
            top_p=top_p,
            temperature=temperature,
            max_tokens=max_tokens,
        )
        buffer = StringIO()
        for chunk in stream:
            if chunk.choices and chunk.choices[0].delta.content:
                buffer.write(chunk.choices[0].delta.content)
        return buffer.getvalue()



from dataclasses import dataclass, field
from pathlib import Path
from xml.etree import ElementTree as ET
from xml.etree.ElementTree import Element

from ..xml.xml import plain_text
from .common import extract_namespace, find_opf_path, strip_namespace
from .zip import Zip


@dataclass
class Toc:
    """
    EPUB 2.0 对应关系:
        - title <-> <navLabel><text>
        - href <-> <content src> (不包含 # 后的部分)
        - fragment <-> <content src> (# 后的部分)
        - children <-> 嵌套的 <navPoint>
        - id <-> <navPoint id>

    EPUB 3.0 对应关系:
        - title <-> <a> 标签的文本内容
        - href <-> <a href> (不包含 # 后的部分)
        - fragment <-> <a href> (# 后的部分)
        - children <-> 嵌套的 <ol><li>
        - id <-> <li id> 或 <a id>
    """

    title: str
    href: str | None = None
    fragment: str | None = None
    id: str | None = None
    children: list["Toc"] = field(default_factory=list)

    @property
    def full_href(self) -> str | None:
        if self.href is None:
            return None
        if self.fragment:
            return f"{self.href}#{self.fragment}"
        return self.href


def read_toc(zip: Zip) -> list[Toc]:
    version = _detect_epub_version(zip)
    toc_path = _find_toc_path(zip, version)

    if toc_path is None:
        return []

    if version == 2:
        return _read_ncx_toc(zip, toc_path)
    else:
        return _read_nav_toc(zip, toc_path)


def write_toc(zip: Zip, toc: list[Toc]) -> None:
    version = _detect_epub_version(zip)
    toc_path = _find_toc_path(zip, version)

    if toc_path is None:
        raise ValueError("Cannot find TOC file in EPUB")

    if version == 2:
        _write_ncx_toc(zip, toc_path, toc)
    else:
        _write_nav_toc(zip, toc_path, toc)


def _detect_epub_version(zip: Zip) -> int:
    opf_path = find_opf_path(zip)
    with zip.read(opf_path) as f:
        content = f.read()
        root = ET.fromstring(content)

        # 检查 package 元素的 version 属性
        version_str = root.get("version", "2.0")

        if version_str.startswith("3"):
            return 3
        else:
            return 2


def _find_toc_path(zip: Zip, version: int) -> Path | None:
    opf_path = find_opf_path(zip)
    opf_dir = opf_path.parent

    with zip.read(opf_path) as f:
        content = f.read()
        root = ET.fromstring(content)
        strip_namespace(root)  # 移除命名空间前缀以简化 XPath

        manifest = root.find(".//manifest")
        if manifest is None:
            return None

        if version == 2:
            # EPUB 2: 查找 NCX 文件 (media-type="application/x-dtbncx+xml")
            for item in manifest.findall("item"):
                media_type = item.get("media-type")
                if media_type == "application/x-dtbncx+xml":
                    href = item.get("href")
                    if href:
                        return opf_dir / href
        else:
            # EPUB 3: 查找 nav 文件 (properties="nav")
            for item in manifest.findall("item"):
                properties = item.get("properties", "")
                if "nav" in properties.split():
                    href = item.get("href")
                    if href:
                        return opf_dir / href

        return None


def _read_ncx_toc(zip: Zip, ncx_path: Path) -> list[Toc]:
    with zip.read(ncx_path) as f:
        content = f.read()
        root = ET.fromstring(content)
        strip_namespace(root)  # 移除命名空间前缀以简化 XPath

        nav_map = root.find(".//navMap")
        if nav_map is None:
            return []

        result = []
        for nav_point in nav_map.findall("navPoint"):
            toc_item = _parse_nav_point(nav_point)
            if toc_item:
                result.append(toc_item)

        return result


def _parse_nav_point(nav_point: Element) -> Toc | None:
    nav_id = nav_point.get("id")
    nav_label = nav_point.find("navLabel")
    if nav_label is None:
        return None

    text_elem = nav_label.find("text")
    if text_elem is None:
        return None

    title = plain_text(text_elem).strip()
    if not title:
        return None

    content_elem = nav_point.find("content")
    href = None
    fragment = None

    if content_elem is not None:
        src = content_elem.get("src")
        if src:
            href, fragment = _split_href(src)

    children = []
    for child_nav_point in nav_point.findall("navPoint"):
        child_toc = _parse_nav_point(child_nav_point)
        if child_toc:
            children.append(child_toc)

    return Toc(
        title=title,
        href=href,
        fragment=fragment,
        id=nav_id,
        children=children,
    )


def _write_ncx_toc(zip: Zip, ncx_path: Path, toc_list: list[Toc]) -> None:
    with zip.read(ncx_path) as f:
        content = f.read()
        root = ET.fromstring(content)
        ns = extract_namespace(root.tag)
        nav_map = root.find(f".//{{{ns}}}navMap" if ns else ".//navMap")
        if nav_map is None:
            raise ValueError("Cannot find navMap in NCX file")
        _update_nav_points(nav_map, toc_list, ns)
        tree = ET.ElementTree(root)
        with zip.replace(ncx_path) as out:
            tree.write(out, encoding="utf-8", xml_declaration=True)


def _update_nav_points(parent: Element, toc_list: list[Toc], ns: str | None, start_play_order: int = 1) -> int:
    tag_prefix = f"{{{ns}}}" if ns else ""
    nav_point_tag = f"{tag_prefix}navPoint"
    existing_nav_points = [elem for elem in parent if elem.tag == nav_point_tag]
    matched_pairs = _match_toc_with_elements(toc_list, existing_nav_points)
    for nav_point in existing_nav_points:
        parent.remove(nav_point)

    play_order = start_play_order
    for toc, existing_elem in matched_pairs:
        if existing_elem is not None:
            nav_point = existing_elem
            _update_nav_point_content(nav_point, toc, ns, play_order)
        else:
            nav_point = _create_nav_point(toc, ns, play_order)

        parent.append(nav_point)
        play_order += 1
        play_order = _update_nav_points(nav_point, toc.children, ns, play_order)

    return play_order


def _update_nav_point_content(nav_point: Element, toc: Toc, ns: str | None, play_order: int) -> None:
    tag_prefix = f"{{{ns}}}" if ns else ""
    if toc.id:
        nav_point.set("id", toc.id)

    nav_point.set("playOrder", str(play_order))

    nav_label = nav_point.find(f"{tag_prefix}navLabel")
    if nav_label is not None:
        text_elem = nav_label.find(f"{tag_prefix}text")
        if text_elem is not None:
            text_elem.text = toc.title

    content_elem = nav_point.find(f"{tag_prefix}content")
    if content_elem is not None and toc.href is not None:
        full_href = toc.full_href
        if full_href:
            content_elem.set("src", full_href)


def _create_nav_point(toc: Toc, ns: str | None, play_order: int) -> Element:
    tag_prefix = f"{{{ns}}}" if ns else ""

    nav_point = Element(f"{tag_prefix}navPoint")
    if toc.id:
        nav_point.set("id", toc.id)
    else:
        nav_point.set("id", f"navPoint-{play_order}")
    nav_point.set("playOrder", str(play_order))

    nav_label = Element(f"{tag_prefix}navLabel")
    text_elem = Element(f"{tag_prefix}text")
    text_elem.text = toc.title
    nav_label.append(text_elem)
    nav_point.append(nav_label)

    if toc.href is not None:
        content_elem = Element(f"{tag_prefix}content")
        full_href = toc.full_href
        if full_href:
            content_elem.set("src", full_href)
        nav_point.append(content_elem)

    return nav_point


def _read_nav_toc(zip: Zip, nav_path: Path) -> list[Toc]:
    with zip.read(nav_path) as f:
        content = f.read()
        root = ET.fromstring(content)

        strip_namespace(root)

        nav_elem = None
        for nav in root.findall(".//nav"):
            epub_type = nav.get("{http://www.idpf.org/2007/ops}type") or nav.get("type")
            if epub_type == "toc":
                nav_elem = nav
                break

        if nav_elem is None:
            return []

        ol = nav_elem.find(".//ol")
        if ol is None:
            return []

        result = []
        for li in ol.findall("li"):
            toc_item = _parse_nav_li(li)
            if toc_item:
                result.append(toc_item)

        return result


def _parse_nav_li(li: Element) -> Toc | None:
    li_id = li.get("id")
    a = li.find("a")
    if a is None:
        span = li.find("span")
        if span is not None:
            title = plain_text(span).strip()
            if not title:
                return None
            href = None
            fragment = None
            a_id = None
        else:
            return None
    else:
        title = plain_text(a).strip()
        if not title:
            return None

        a_id = a.get("id")
        href_attr = a.get("href")

        if href_attr:
            href, fragment = _split_href(href_attr)
        else:
            href = None
            fragment = None

    final_id = li_id if li_id else (a_id if "a_id" in locals() else None)
    children = []
    child_ol = li.find("ol")
    if child_ol is not None:
        for child_li in child_ol.findall("li"):
            child_toc = _parse_nav_li(child_li)
            if child_toc:
                children.append(child_toc)

    return Toc(
        title=title,
        href=href,
        fragment=fragment,
        id=final_id,
        children=children,
    )


def _write_nav_toc(zip: Zip, nav_path: Path, toc_list: list[Toc]) -> None:
    with zip.read(nav_path) as f:
        content = f.read()
        root = ET.fromstring(content)
        ns = extract_namespace(root.tag)
        nav_elem = None
        for nav in root.findall(f".//{{{ns}}}nav" if ns else ".//nav"):
            epub_type = nav.get("{http://www.idpf.org/2007/ops}type") or nav.get("type") or nav.get(f"{{{ns}}}type")
            if epub_type == "toc":
                nav_elem = nav
                break

        if nav_elem is None:
            raise ValueError("Cannot find nav element with type='toc'")

        ol = nav_elem.find(f".//{{{ns}}}ol" if ns else ".//ol")
        if ol is None:
            raise ValueError("Cannot find ol in nav element")

        _update_nav_lis(ol, toc_list, ns)

        tree = ET.ElementTree(root)
        with zip.replace(nav_path) as out:
            tree.write(out, encoding="utf-8", xml_declaration=True)


def _update_nav_lis(ol: Element, toc_list: list[Toc], ns: str | None) -> None:
    tag_prefix = f"{{{ns}}}" if ns else ""
    li_tag = f"{tag_prefix}li"
    existing_lis = [elem for elem in ol if elem.tag == li_tag]
    matched_pairs = _match_toc_with_elements(toc_list, existing_lis)

    for li in existing_lis:
        ol.remove(li)

    for toc, existing_elem in matched_pairs:
        if existing_elem is not None:
            li = existing_elem
            _update_nav_li_content(li, toc, ns)
        else:
            li = _create_nav_li(toc, ns)

        ol.append(li)

        if toc.children:
            child_ol = li.find(f"{tag_prefix}ol")
            if child_ol is None:
                child_ol = Element(f"{tag_prefix}ol")
                li.append(child_ol)
            _update_nav_lis(child_ol, toc.children, ns)


def _update_nav_li_content(li: Element, toc: Toc, ns: str | None) -> None:
    tag_prefix = f"{{{ns}}}" if ns else ""
    if toc.id:
        li.set("id", toc.id)

    a = li.find(f"{tag_prefix}a")
    span = li.find(f"{tag_prefix}span")

    if toc.href is not None:
        if a is not None:
            a.text = toc.title
            full_href = toc.full_href
            if full_href:
                a.set("href", full_href)
        elif span is not None:
            li.remove(span)
            a = Element(f"{tag_prefix}a")
            a.text = toc.title
            full_href = toc.full_href
            if full_href:
                a.set("href", full_href)
            li.insert(0, a)
    else:
        if span is not None:
            span.text = toc.title
        elif a is not None:
            li.remove(a)
            span = Element(f"{tag_prefix}span")
            span.text = toc.title
            li.insert(0, span)


def _create_nav_li(toc: Toc, ns: str | None) -> Element:
    tag_prefix = f"{{{ns}}}" if ns else ""
    li = Element(f"{tag_prefix}li")

    if toc.id:
        li.set("id", toc.id)

    if toc.href is not None:
        a = Element(f"{tag_prefix}a")
        a.text = toc.title
        full_href = toc.full_href
        if full_href:
            a.set("href", full_href)
        li.append(a)
    else:
        span = Element(f"{tag_prefix}span")
        span.text = toc.title
        li.append(span)

    return li


def _split_href(href: str) -> tuple[str | None, str | None]:
    if "#" in href:
        parts = href.split("#", 1)
        return parts[0] if parts[0] else None, parts[1] if parts[1] else None
    else:
        return href, None


def _match_toc_with_elements(toc_list: list[Toc], elements: list[Element]) -> list[tuple[Toc, Element | None]]:
    """
    使用混合策略匹配 Toc 对象和 XML 元素

    策略优先级：
    1. 通过 id 匹配
    2. 通过 href 匹配
    3. 通过位置匹配
    """
    result = []
    used_elements = set()

    for toc in toc_list:
        matched = None
        if toc.id:
            for i, elem in enumerate(elements):
                if i in used_elements:
                    continue
                elem_id = elem.get("id")
                if elem_id == toc.id:
                    matched = elem
                    used_elements.add(i)
                    break
        result.append((toc, matched))

    for i, (toc, matched) in enumerate(result):
        if matched is None and toc.href:
            for j, elem in enumerate(elements):
                if j in used_elements:
                    continue
                elem_href = _extract_href_from_element(elem)
                if elem_href and elem_href == toc.full_href:
                    result[i] = (toc, elem)
                    used_elements.add(j)
                    break

    unmatched_indices = [i for i, (_, matched) in enumerate(result) if matched is None]
    available_elements = [elem for j, elem in enumerate(elements) if j not in used_elements]

    for i, elem in zip(unmatched_indices, available_elements):
        toc, _ = result[i]
        result[i] = (toc, elem)

    return result


def _extract_href_from_element(elem: Element) -> str | None:
    # NCX 格式：查找 content/@src
    content = elem.find(".//content")
    if content is not None:
        return content.get("src")

    # nav 格式：查找 a/@href
    a = elem.find(".//a")
    if a is not None:
        return a.get("href")

    return None



from collections.abc import Callable
from pathlib import Path
from xml.etree.ElementTree import Element

from .epub import Placeholder, Zip, is_placeholder_tag, read_toc, search_spine_paths, write_toc
from .epub.common import find_opf_path
from .llm import LLM
from .xml import XMLLikeNode, deduplicate_ids_in_element, find_first, plain_text
from .xml_translator import XMLGroupContext, XMLTranslator, submit_text_segments


def translate(
    llm: LLM,
    source_path: Path,
    target_path: Path,
    target_language: str,
    user_prompt: str | None = None,
    max_retries: int = 5,
    max_group_tokens: int = 1200,
    on_progress: Callable[[float], None] | None = None,
) -> None:
    translator = XMLTranslator(
        llm=llm,
        target_language=target_language,
        user_prompt=user_prompt,
        ignore_translated_error=False,
        max_retries=max_retries,
        max_fill_displaying_errors=10,
        group_context=XMLGroupContext(
            encoding=llm.encoding,
            max_group_tokens=max_group_tokens,
        ),
    )
    with Zip(source_path, target_path) as zip:
        # Progress distribution: TOC 3%, metadata 2%, chapters 95%
        TOC_PROGRESS = 0.03
        METADATA_PROGRESS = 0.02
        CHAPTERS_PROGRESS = 0.95

        # Count total chapters for progress calculation (lightweight, no content loading)
        total_chapters = _count_chapters(zip)
        chapter_progress_step = CHAPTERS_PROGRESS / total_chapters if total_chapters > 0 else 0

        current_progress = 0.0

        # Translate TOC
        _translate_toc(translator, zip)
        current_progress += TOC_PROGRESS
        if on_progress:
            on_progress(current_progress)

        # Translate metadata
        _translate_metadata(translator, zip)
        current_progress += METADATA_PROGRESS
        if on_progress:
            on_progress(current_progress)

        # Translate chapters
        processed_chapters = 0
        for element, text_segments, (chapter_path, xml, placeholder) in translator.translate_to_text_segments(
            items=_search_chapter_items(zip),
        ):
            submit_text_segments(
                element=element,
                text_segments=(
                    segment
                    for segment in text_segments
                    if not any(is_placeholder_tag(e.tag) for e in segment.parent_stack)
                ),
            )
            placeholder.recover()
            deduplicate_ids_in_element(xml.element)
            with zip.replace(chapter_path) as target_file:
                xml.save(target_file, is_html_like=True)

            # Update progress after each chapter
            processed_chapters += 1
            current_progress = TOC_PROGRESS + METADATA_PROGRESS + (processed_chapters * chapter_progress_step)
            if on_progress:
                on_progress(current_progress)


def _translate_toc(translator: XMLTranslator, zip: Zip):
    """Translate TOC (Table of Contents) titles."""
    toc_list = read_toc(zip)
    if not toc_list:
        return

    # Collect all titles recursively
    titles_to_translate: list[str] = []

    def collect_titles(items):
        for item in items:
            titles_to_translate.append(item.title)
            if item.children:
                collect_titles(item.children)

    collect_titles(toc_list)

    # Create XML elements for translation
    elements_to_translate = Element("toc")
    elements_to_translate.extend(_create_text_element(title) for title in titles_to_translate)

    # Translate all titles at once
    translated_element = translator.translate_to_element(elements_to_translate)

    # Extract translated texts
    from builtins import zip as builtin_zip

    translated_titles = [
        plain_text(elem) if elem is not None else original
        for elem, original in builtin_zip(translated_element, titles_to_translate)
    ]

    # Fill back translated titles
    title_index = 0

    def fill_titles(items):
        nonlocal title_index
        for item in items:
            item.title = translated_titles[title_index]
            title_index += 1
            if item.children:
                fill_titles(item.children)

    fill_titles(toc_list)

    # Write back the translated TOC
    write_toc(zip, toc_list)


def _translate_metadata(translator: XMLTranslator, zip: Zip):
    """Translate metadata fields in OPF file."""
    opf_path = find_opf_path(zip)

    with zip.read(opf_path) as f:
        xml = XMLLikeNode(f)

    # Find metadata element
    metadata_elem = None
    for child in xml.element:
        if child.tag.endswith("metadata"):
            metadata_elem = child
            break

    if metadata_elem is None:
        return

    # Collect metadata fields to translate
    # Skip fields that should not be translated
    skip_fields = {
        "language",
        "identifier",
        "date",
        "meta",
        "contributor",  # Usually technical information
    }

    fields_to_translate: list[tuple[Element, str]] = []

    for elem in metadata_elem:
        # Get tag name without namespace
        tag_name = elem.tag.split("}")[-1] if "}" in elem.tag else elem.tag

        # Check if element has text content and should be translated
        if elem.text and elem.text.strip() and tag_name not in skip_fields:
            fields_to_translate.append((elem, elem.text.strip()))

    if not fields_to_translate:
        return

    # Create XML elements for translation
    elements_to_translate = Element("metadata")
    elements_to_translate.extend(_create_text_element(text) for _, text in fields_to_translate)

    # Translate all metadata at once
    translated_element = translator.translate_to_element(elements_to_translate)

    # Fill back translated texts
    from builtins import zip as builtin_zip

    for (elem, _), translated_elem in builtin_zip(fields_to_translate, translated_element, strict=True):
        if translated_elem is not None:
            translated_text = plain_text(translated_elem)
            if translated_text:
                elem.text = translated_text

    # Write back the modified OPF file
    with zip.replace(opf_path) as f:
        xml.save(f)


def _count_chapters(zip: Zip) -> int:
    """Count total chapters without loading content (lightweight)."""
    return sum(1 for _ in search_spine_paths(zip))


def _search_chapter_items(zip: Zip):
    for chapter_path in search_spine_paths(zip):
        with zip.read(chapter_path) as chapter_file:
            xml = XMLLikeNode(chapter_file)
        body_element = find_first(xml.element, "body")
        if body_element is not None:
            placeholder = Placeholder(body_element)
            yield body_element, (chapter_path, xml, placeholder)


def _create_text_element(text: str) -> Element:
    elem = Element("text")
    elem.text = text
    return elem



from xml.etree.ElementTree import Element

from .xml import iter_with_stack

_ID_KEY = "id"
_SUFFIX = "__translated"


def deduplicate_ids_in_element(element: Element) -> Element:
    seen_ids: set[str] = set()
    original_id_count: dict[str, int] = {}

    for _, sub_element in iter_with_stack(element):
        if _ID_KEY not in sub_element.attrib:
            continue
        original_id = sub_element.attrib[_ID_KEY]

        if original_id not in seen_ids:
            seen_ids.add(original_id)
            original_id_count[original_id] = 1
        else:
            original_id_count[original_id] = original_id_count.get(original_id, 1) + 1
            occurrence = original_id_count[original_id]

            if occurrence == 2:
                new_id = f"{original_id}{_SUFFIX}"
            else:
                new_id = f"{original_id}{_SUFFIX}_{occurrence - 1}"

            counter = occurrence - 1
            while new_id in seen_ids:
                counter += 1
                new_id = f"{original_id}{_SUFFIX}_{counter}"

            sub_element.attrib["id"] = new_id
            seen_ids.add(new_id)

    return element



from collections.abc import Callable, Generator, Iterable

from .chunk import split_into_chunks
from .segment import ST, T


def split(
    segments: Iterable[ST],
    transform: Callable[[list[ST]], list[T]],
    max_group_tokens: int,
) -> Generator[T, None, None]:
    for group in split_into_chunks(segments, max_group_tokens):
        head = list(
            _truncate_extra_content(
                segments=group.head,
                remain_left=False,
                remain_tokens=group.head_remain_tokens,
            )
        )
        tail = list(
            _truncate_extra_content(
                segments=group.tail,
                remain_left=True,
                remain_tokens=group.tail_remain_tokens,
            )
        )
        transformed = transform(head + group.body + tail)

        if len(tail) > 0:  # 避免 target[N:-0] 切片错误
            yield from transformed[len(head) : -len(tail)]
        else:
            yield from transformed[len(head) :]


def _truncate_extra_content(segments: list[ST], remain_left: bool, remain_tokens: int):
    tokens_list: list[int] = [segment.tokens for segment in segments]
    segments = list(segments)
    for tokens in tokens_list if remain_left else reversed(tokens_list):
        if remain_tokens <= 0:
            break
        next_segment = segments.pop(0) if remain_left else segments.pop()
        if remain_tokens < tokens:
            if remain_left:
                next_segment = next_segment.truncate_after_head(remain_tokens)
            else:
                next_segment = next_segment.truncate_before_tail(remain_tokens)
            remain_tokens = 0
        else:
            remain_tokens -= tokens
        yield next_segment



import datetime
import hashlib
import json
from collections.abc import Callable, Generator
from importlib.resources import files
from logging import DEBUG, FileHandler, Formatter, Logger, getLogger
from os import PathLike
from pathlib import Path

from jinja2 import Environment, Template
from tiktoken import Encoding, get_encoding

from ..template import create_env
from .executor import LLMExecutor
from .increasable import Increasable
from .types import Message, MessageRole, R


class LLM:
    def __init__(
        self,
        key: str,
        url: str,
        model: str,
        token_encoding: str,
        cache_path: PathLike | None = None,
        timeout: float | None = None,
        top_p: float | tuple[float, float] | None = None,
        temperature: float | tuple[float, float] | None = None,
        retry_times: int = 5,
        retry_interval_seconds: float = 6.0,
        log_dir_path: PathLike | None = None,
    ):
        prompts_path = Path(str(files("epub_translator"))) / "data"
        self._templates: dict[str, Template] = {}
        self._encoding: Encoding = get_encoding(token_encoding)
        self._env: Environment = create_env(prompts_path)
        self._logger_save_path: Path | None = None
        self._cache_path: Path | None = None

        if cache_path is not None:
            self._cache_path = Path(cache_path)
            if not self._cache_path.exists():
                self._cache_path.mkdir(parents=True, exist_ok=True)
            elif not self._cache_path.is_dir():
                self._cache_path = None

        if log_dir_path is not None:
            self._logger_save_path = Path(log_dir_path)
            if not self._logger_save_path.exists():
                self._logger_save_path.mkdir(parents=True, exist_ok=True)
            elif not self._logger_save_path.is_dir():
                self._logger_save_path = None

        self._executor = LLMExecutor(
            url=url,
            model=model,
            api_key=key,
            timeout=timeout,
            top_p=Increasable(top_p),
            temperature=Increasable(temperature),
            retry_times=retry_times,
            retry_interval_seconds=retry_interval_seconds,
            create_logger=self._create_logger,
        )

    @property
    def encoding(self) -> Encoding:
        return self._encoding

    def request(
        self,
        input: str | list[Message],
        parser: Callable[[str], R] = lambda x: x,
        max_tokens: int | None = None,
    ) -> R:
        messages: list[Message]
        if isinstance(input, str):
            messages = [Message(role=MessageRole.USER, message=input)]
        else:
            messages = input

        # Check cache if cache_path is set
        if self._cache_path is not None:
            cache_key = self._compute_messages_hash(messages)
            cache_file = self._cache_path / f"{cache_key}.txt"

            if cache_file.exists():
                cached_content = cache_file.read_text(encoding="utf-8")
                return parser(cached_content)

        # Make the actual request
        response = self._executor.request(
            messages=messages,
            parser=lambda x: x,
            max_tokens=max_tokens,
        )

        # Save to cache if cache_path is set
        if self._cache_path is not None:
            cache_key = self._compute_messages_hash(messages)
            cache_file = self._cache_path / f"{cache_key}.txt"
            cache_file.write_text(response, encoding="utf-8")

        return parser(response)

    def template(self, template_name: str) -> Template:
        template = self._templates.get(template_name, None)
        if template is None:
            template = self._env.get_template(template_name)
            self._templates[template_name] = template
        return template

    def _compute_messages_hash(self, messages: list[Message]) -> str:
        """Compute SHA-512 hash of m·essages for cache key."""
        messages_dict = [{"role": msg.role.value, "message": msg.message} for msg in messages]
        messages_json = json.dumps(messages_dict, ensure_ascii=False, sort_keys=True)
        return hashlib.sha512(messages_json.encode("utf-8")).hexdigest()

    def _create_logger(self) -> Logger | None:
        if self._logger_save_path is None:
            return None

        now = datetime.datetime.now(datetime.timezone.utc)
        timestamp = now.strftime("%Y-%m-%d %H-%M-%S %f")
        file_path = self._logger_save_path / f"request {timestamp}.log"
        logger = getLogger(f"LLM Request {timestamp}")
        logger.setLevel(DEBUG)
        handler = FileHandler(file_path, encoding="utf-8")
        handler.setLevel(DEBUG)
        handler.setFormatter(Formatter("%(asctime)s    %(message)s", "%H:%M:%S"))
        logger.addHandler(handler)

        return logger

    def _search_quotes(self, kind: str, response: str) -> Generator[str, None, None]:
        start_marker = f"```{kind}"
        end_marker = "```"
        start_index = 0

        while True:
            start_index = self._find_ignore_case(
                raw=response,
                sub=start_marker,
                start=start_index,
            )
            if start_index == -1:
                break

            end_index = self._find_ignore_case(
                raw=response,
                sub=end_marker,
                start=start_index + len(start_marker),
            )
            if end_index == -1:
                break

            extracted_text = response[start_index + len(start_marker) : end_index].strip()
            yield extracted_text
            start_index = end_index + len(end_marker)

    def _find_ignore_case(self, raw: str, sub: str, start: int = 0):
        if not sub:
            return 0 if 0 >= start else -1

        raw_len, sub_len = len(raw), len(sub)
        for i in range(start, raw_len - sub_len + 1):
            match = True
            for j in range(sub_len):
                if raw[i + j].lower() != sub[j].lower():
                    match = False
                    break
            if match:
                return i
        return -1



from collections.abc import Generator, Iterable
from dataclasses import dataclass
from typing import Generic

from resource_segmentation import Resource, Segment, split

from .segment import ST

_INCISION = 0


@dataclass
class Chunk(Generic[ST]):
    head_remain_tokens: int
    tail_remain_tokens: int
    head: list[ST]
    body: list[ST]
    tail: list[ST]


def split_into_chunks(segments: Iterable[ST], max_group_tokens: int) -> Generator[Chunk[ST], None, None]:
    for group in split(
        max_segment_count=max_group_tokens,
        gap_rate=0.07,
        tail_rate=0.5,
        border_incision=_INCISION,
        resources=(
            Resource(
                count=segment.tokens,
                start_incision=_INCISION,
                end_incision=_INCISION,
                payload=segment,
            )
            for segment in segments
        ),
    ):
        yield Chunk(
            head_remain_tokens=group.head_remain_count,
            tail_remain_tokens=group.tail_remain_count,
            head=list(_expand_payloads(group.head)),
            body=list(_expand_payloads(group.body)),
            tail=list(_expand_payloads(group.tail)),
        )


def _expand_payloads(target: list[Resource[ST] | Segment[ST]]) -> Generator[ST, None, None]:
    for item in target:
        if isinstance(item, Resource):
            yield item.payload
        elif isinstance(item, Segment):
            for resource in item.resources:
                yield resource.payload



from xml.etree.ElementTree import Element

from .tag import Tag, TagKind


def tag_to_element(tag: Tag) -> Element:
    element = Element(tag.name)
    for attr_name, attr_value in tag.attributes:
        element.set(attr_name, attr_value)
    return element


def element_to_tag(element: Element, kind: TagKind, proto: str = "") -> Tag:
    tag = Tag(
        kind=kind,
        name=element.tag,
        proto=proto,
        attributes=[],
    )
    if kind != TagKind.CLOSING:
        for attr_name in sorted(list(element.keys())):
            attr_value = element.get(attr_name, "")
            tag.attributes.append((attr_name, attr_value))

    # To make LLM easier to understand, the naming here is restricted in a more strict way.
    # https://github.com/oomol-lab/pdf-craft/issues/149
    invalid_name = tag.find_invalid_name()
    if invalid_name is not None:
        raise ValueError(f"find invalid tag name or attribute name: {invalid_name}")

    invalid_attr_pair = tag.find_invalid_attr_value()
    if invalid_attr_pair is not None:
        attr_name, attr_value = invalid_attr_pair
        raise ValueError(f'find invalid attribute value: {attr_name}="{attr_value}"')

    return tag



import io
import re
from typing import IO
from xml.etree.ElementTree import Element, fromstring, tostring

from .xml import iter_with_stack

_COMMON_NAMESPACES = {
    "http://www.w3.org/1999/xhtml": "xhtml",
    "http://www.idpf.org/2007/ops": "epub",
    "http://www.w3.org/1998/Math/MathML": "m",
    "http://purl.org/dc/elements/1.1/": "dc",
    "http://www.daisy.org/z3986/2005/ncx/": "ncx",
    "http://www.idpf.org/2007/opf": "opf",
    "http://www.w3.org/2000/svg": "svg",
    "urn:oasis:names:tc:opendocument:xmlns:container": "container",
}

_ROOT_NAMESPACES = {
    "http://www.w3.org/1999/xhtml",  # XHTML
    "http://www.daisy.org/z3986/2005/ncx/",  # NCX
    "http://www.idpf.org/2007/opf",  # OPF
    "urn:oasis:names:tc:opendocument:xmlns:container",  # Container
}

_ENCODING_PATTERN = re.compile(r'encoding\s*=\s*["\']([^"\']+)["\']', re.IGNORECASE)
_FIRST_ELEMENT_PATTERN = re.compile(r"<(?![?!])[a-zA-Z]")
_NAMESPACE_IN_TAG = re.compile(r"\{([^}]+)\}")

# HTML 规定了一系列自闭标签，这些标签需要改成非自闭的，因为 EPub 格式不支持
# https://www.tutorialspoint.com/which-html-tags-are-self-closing
_EMPTY_TAGS = (
    "br",
    "hr",
    "input",
    "col",
    "base",
    "meta",
    "area",
)

_EMPTY_TAG_PATTERN = re.compile(r"<(" + "|".join(_EMPTY_TAGS) + r")(\s[^>]*?)\s*/?>")


class XMLLikeNode:
    def __init__(self, file: IO[bytes]) -> None:
        raw_content = file.read()
        self._encoding: str = _detect_encoding(raw_content)
        content = raw_content.decode(self._encoding)
        self._header, xml_content = _extract_header(content)
        try:
            self.element = fromstring(xml_content)
        except Exception as error:
            raise ValueError("Failed to parse XML-like content") from error
        self._namespaces: dict[str, str] = _extract_and_clean_namespaces(self.element)

    @property
    def encoding(self) -> str:
        return self._encoding

    @property
    def namespaces(self) -> list[str]:
        return list(self._namespaces.keys())

    def save(self, file: IO[bytes], is_html_like: bool = False) -> None:
        writer = io.TextIOWrapper(file, encoding=self._encoding, write_through=True)
        try:
            if self._header:
                writer.write(self._header)

            content = _serialize_with_namespaces(element=self.element, namespaces=self._namespaces)
            if is_html_like:
                content = re.sub(
                    pattern=_EMPTY_TAG_PATTERN,
                    repl=lambda m: f"<{m.group(1)}{m.group(2)}>",
                    string=content,
                )
            else:
                content = re.sub(
                    pattern=_EMPTY_TAG_PATTERN,
                    repl=lambda m: f"<{m.group(1)}{m.group(2)} />",
                    string=content,
                )
            writer.write(content)

        finally:
            writer.detach()


def _detect_encoding(raw_content: bytes) -> str:
    if raw_content.startswith(b"\xef\xbb\xbf"):
        return "utf-8-sig"
    elif raw_content.startswith(b"\xff\xfe"):
        return "utf-16-le"
    elif raw_content.startswith(b"\xfe\xff"):
        return "utf-16-be"

    # 尝试从 XML 声明中提取编码：只读取前 1024 字节来查找 XML 声明
    header_bytes = raw_content[:1024]
    for try_encoding in ("utf-8", "utf-16-le", "utf-16-be", "iso-8859-1"):
        try:
            header_str = header_bytes.decode(try_encoding)
            match = _ENCODING_PATTERN.search(header_str)
            if match:
                declared_encoding = match.group(1).lower()
                try:
                    raw_content.decode(declared_encoding)
                    return declared_encoding
                except (LookupError, UnicodeDecodeError):
                    pass
        except UnicodeDecodeError:
            continue

    try:
        raw_content.decode("utf-8")
        return "utf-8"
    except UnicodeDecodeError:
        pass
    return "iso-8859-1"


def _extract_header(content: str) -> tuple[str, str]:
    match = _FIRST_ELEMENT_PATTERN.search(content)
    if match:
        split_pos = match.start()
        header = content[:split_pos]
        xml_content = content[split_pos:]
        return header, xml_content
    return "", content


def _extract_and_clean_namespaces(element: Element):
    namespaces: dict[str, str] = {}
    for _, elem in iter_with_stack(element):
        match = _NAMESPACE_IN_TAG.match(elem.tag)
        if match:
            namespace_uri = match.group(1)
            if namespace_uri not in namespaces:
                prefix = _COMMON_NAMESPACES.get(namespace_uri, f"ns{len(namespaces)}")
                namespaces[namespace_uri] = prefix

            tag_name = elem.tag[len(match.group(0)) :]
            elem.tag = tag_name

        for attr_key in list(elem.attrib.keys()):
            match = _NAMESPACE_IN_TAG.match(attr_key)
            if match:
                namespace_uri = match.group(1)
                if namespace_uri not in namespaces:
                    prefix = _COMMON_NAMESPACES.get(namespace_uri, f"ns{len(namespaces)}")
                    namespaces[namespace_uri] = prefix

                attr_name = attr_key[len(match.group(0)) :]
                attr_value = elem.attrib.pop(attr_key)
                elem.attrib[attr_name] = attr_value
    return namespaces


def _serialize_with_namespaces(
    element: Element,
    namespaces: dict[str, str],
) -> str:
    for namespace_uri, prefix in namespaces.items():
        if namespace_uri in _ROOT_NAMESPACES:
            element.attrib["xmlns"] = namespace_uri
        else:
            element.attrib[f"xmlns:{prefix}"] = namespace_uri
    xml_string = tostring(element, encoding="unicode")
    for namespace_uri, prefix in namespaces.items():
        if namespace_uri in _ROOT_NAMESPACES:
            xml_string = xml_string.replace(f"{{{namespace_uri}}}", "")
        else:
            xml_string = xml_string.replace(f"{{{namespace_uri}}}", f"{prefix}:")
        pattern = r'\s+xmlns:(ns\d+)="' + re.escape(namespace_uri) + r'"'
        xml_string = re.sub(pattern, "", xml_string)
    return xml_string



from collections.abc import Generator
from pathlib import Path
from xml.etree import ElementTree as ET

from .common import find_opf_path, strip_namespace
from .zip import Zip


def search_spine_paths(zip: Zip) -> Generator[Path, None, None]:
    opf_path = find_opf_path(zip)
    opf_dir = opf_path.parent

    with zip.read(opf_path) as f:
        content = f.read()
        root = ET.fromstring(content)
        strip_namespace(root)

        manifest = root.find(".//manifest")
        if manifest is None:
            return

        manifest_items = {}
        for item in manifest.findall("item"):
            item_id = item.get("id")
            item_href = item.get("href")
            media_type = item.get("media-type", "")
            if item_id and item_href:
                manifest_items[item_id] = (item_href, media_type)

        spine = root.find(".//spine")
        if spine is None:
            return

        for itemref in spine.findall("itemref"):
            idref = itemref.get("idref")
            if not idref:
                continue

            if idref in manifest_items:
                href, media_type = manifest_items[idref]
                if media_type in ("application/xhtml+xml", "text/html"):
                    yield opf_dir / href



from collections.abc import Generator, Iterable
from enum import Enum, auto
from io import StringIO

from .tag import Tag, TagKind, is_valid_name_char, is_valid_value_char

_SPACES = (" ", "\n")


class _Phase(Enum):
    OUTSIDE = auto()
    LEFT_BRACKET = auto()
    LEFT_SLASH = auto()
    TAG_NAME = auto()
    TAG_GAP = auto()
    ATTRIBUTE_NAME = auto()
    ATTRIBUTE_NAME_EQUAL = auto()
    ATTRIBUTE_VALUE = auto()
    MUST_CLOSING_SIGN = auto()


class _ParsedResult(Enum):
    Continue = auto()
    Success = auto()
    Failed = auto()


def parse_tags(chars: Iterable[str]) -> Generator[str | Tag, None, None]:
    yield from _XMLTagsParser().do(chars)


class _XMLTagsParser:
    def __init__(self):
        self._outside_buffer: StringIO = StringIO()
        self._tag_buffer: StringIO = StringIO()
        self._tag: Tag | None = None
        self._phase: _Phase = _Phase.OUTSIDE

    def do(self, chars: Iterable[str]) -> Generator[str | Tag, None, None]:
        for char in chars:
            parsed_result = self._parse_char(char)
            yield from self._generate_by_result(parsed_result)

        self._outside_buffer.write(self._tag_buffer.getvalue())
        outside_text = self._outside_buffer.getvalue()
        if outside_text != "":
            yield outside_text

    def _parse_char(self, char: str) -> _ParsedResult:
        parsed_result: _ParsedResult = _ParsedResult.Continue

        if self._phase == _Phase.OUTSIDE:
            if char != "<":
                self._outside_buffer.write(char)
            else:
                self._phase = _Phase.LEFT_BRACKET
                self._tag_buffer.write(char)
                self._tag = Tag(
                    kind=TagKind.OPENING,
                    name="",
                    proto="",
                    attributes=[],
                )
        else:
            assert self._tag is not None
            self._tag_buffer.write(char)

            if self._phase == _Phase.LEFT_BRACKET:
                if char == "/":
                    self._tag.kind = TagKind.CLOSING
                    self._phase = _Phase.LEFT_SLASH
                elif is_valid_name_char(char):
                    self._tag.name += char
                    self._phase = _Phase.TAG_NAME
                else:
                    parsed_result = _ParsedResult.Failed

            elif self._phase == _Phase.LEFT_SLASH:
                if is_valid_name_char(char):
                    self._tag.name += char
                    self._phase = _Phase.TAG_NAME
                else:
                    parsed_result = _ParsedResult.Failed

            elif self._phase == _Phase.TAG_NAME:
                if char in _SPACES:
                    self._phase = _Phase.TAG_GAP
                elif is_valid_name_char(char):
                    self._tag.name += char
                elif char == ">":
                    parsed_result = _ParsedResult.Success
                elif char == "/" and self._tag.kind == TagKind.OPENING:
                    self._tag.kind = TagKind.SELF_CLOSING
                    self._phase = _Phase.MUST_CLOSING_SIGN
                else:
                    parsed_result = _ParsedResult.Failed

            elif self._phase == _Phase.TAG_GAP:
                if char in _SPACES:
                    pass
                elif is_valid_name_char(char):
                    self._tag.attributes.append((char, ""))
                    self._phase = _Phase.ATTRIBUTE_NAME
                elif char == ">":
                    parsed_result = _ParsedResult.Success
                elif char == "/" and self._tag.kind == TagKind.OPENING:
                    self._tag.kind = TagKind.SELF_CLOSING
                    self._phase = _Phase.MUST_CLOSING_SIGN
                else:
                    parsed_result = _ParsedResult.Failed

            elif self._phase == _Phase.ATTRIBUTE_NAME:
                if is_valid_name_char(char):
                    attr_name, attr_value = self._tag.attributes[-1]
                    attr_name = attr_name + char
                    self._tag.attributes[-1] = (attr_name, attr_value)
                elif char == "=":
                    self._phase = _Phase.ATTRIBUTE_NAME_EQUAL
                else:
                    parsed_result = _ParsedResult.Failed

            elif self._phase == _Phase.ATTRIBUTE_NAME_EQUAL:
                if char == '"':
                    self._phase = _Phase.ATTRIBUTE_VALUE
                else:
                    parsed_result = _ParsedResult.Failed

            elif self._phase == _Phase.ATTRIBUTE_VALUE:
                if is_valid_value_char(char):
                    attr_name, attr_value = self._tag.attributes[-1]
                    attr_value = attr_value + char
                    self._tag.attributes[-1] = (attr_name, attr_value)
                elif char == '"':
                    self._phase = _Phase.TAG_GAP
                else:
                    parsed_result = _ParsedResult.Failed

            elif self._phase == _Phase.MUST_CLOSING_SIGN:
                if char == ">":
                    parsed_result = _ParsedResult.Success
                else:
                    parsed_result = _ParsedResult.Failed

        return parsed_result

    def _generate_by_result(self, parsed_result: _ParsedResult) -> Generator[str | Tag, None, None]:
        if parsed_result == _ParsedResult.Success:
            assert self._tag is not None
            if self._is_tag_valid(self._tag):
                outside_text = self._outside_buffer.getvalue()
                self._clear_buffer(self._outside_buffer)
                self._clear_buffer(self._tag_buffer)
                if outside_text != "":
                    yield outside_text
                yield self._tag
            else:
                self._tag.proto = self._tag_buffer.getvalue()
                self._outside_buffer.write(self._tag.proto)
                self._clear_buffer(self._tag_buffer)
            self._tag = None
            self._phase = _Phase.OUTSIDE

        elif parsed_result == _ParsedResult.Failed:
            self._outside_buffer.write(self._tag_buffer.getvalue())
            self._clear_buffer(self._tag_buffer)
            self._phase = _Phase.OUTSIDE

    def _is_tag_valid(self, tag: Tag) -> bool:
        if tag.kind == TagKind.CLOSING and len(tag.attributes) > 0:
            return False
        if tag.find_invalid_name() is not None:
            return False
        return True

    def _clear_buffer(self, buffer: StringIO):
        buffer.truncate(0)
        buffer.seek(0)



from collections.abc import Generator, Iterable, Iterator
from enum import Enum, auto
from xml.etree.ElementTree import Element

from tiktoken import Encoding

from .utils import expand_left_element_texts, expand_right_element_texts, normalize_text_in_element


def group_fragmented_elements(
    encoding: Encoding,
    elements: Iterable[Element],
    group_max_tokens: int,
) -> Generator[list[Element], None, None]:
    remain_tokens_count: int = group_max_tokens
    elements_buffer: list[Element] = []

    for element in elements:
        if remain_tokens_count <= 0:
            remain_tokens_count = group_max_tokens
            if elements_buffer:
                yield elements_buffer
                elements_buffer = []

        counter = _XMLCounter(encoding, element)
        cost_tokens_count = counter.advance_tokens(remain_tokens_count)
        remain_tokens_count -= cost_tokens_count
        if not counter.can_advance():
            elements_buffer.append(element)
            continue

        if elements_buffer:
            yield elements_buffer
            elements_buffer = []

        remain_tokens_count = group_max_tokens - cost_tokens_count
        cost_tokens_count = counter.advance_tokens(remain_tokens_count)
        if not counter.can_advance():
            elements_buffer.append(element)
            remain_tokens_count -= cost_tokens_count
            continue

        remain_tokens_count = group_max_tokens
        yield [element]

    if elements_buffer:
        yield elements_buffer


class _TextItemKind(Enum):
    TEXT = auto()
    XML_TAG = auto()


class _XMLCounter:
    def __init__(self, encoding: Encoding, root: Element) -> None:
        self._encoding: Encoding = encoding
        self._text_iter: Iterator[str] = iter(self._expand_texts(root))
        self._remain_tokens_count: int = 0
        self._next_text_buffer: str | None = None

    def can_advance(self) -> bool:
        if self._remain_tokens_count > 0:
            return True
        if self._next_text_buffer is None:
            self._next_text_buffer = next(self._text_iter, None)
        return self._next_text_buffer is not None

    def _expand_texts(self, element: Element) -> Generator[str, None, None]:
        xml_tags_buffer: list[str] = []  # 这类过于碎片化，需拼凑避免 encoding 失效
        for kind, text in self._expand_text_items(element):
            if kind == _TextItemKind.XML_TAG:
                xml_tags_buffer.append(text)
            elif kind == _TextItemKind.TEXT:
                if xml_tags_buffer:
                    yield "".join(xml_tags_buffer)
                    xml_tags_buffer = []
                yield text
        if xml_tags_buffer:
            yield "".join(xml_tags_buffer)

    def _expand_text_items(self, element: Element) -> Generator[tuple[_TextItemKind, str], None, None]:
        for text in expand_left_element_texts(element):
            yield _TextItemKind.XML_TAG, text

        text = normalize_text_in_element(element.text)
        if text is not None:
            yield _TextItemKind.TEXT, text
        for child in element:
            yield from self._expand_text_items(child)
            tail = normalize_text_in_element(child.tail)
            if tail is not None:
                yield _TextItemKind.TEXT, tail

        for text in expand_right_element_texts(element):
            yield _TextItemKind.XML_TAG, text

    def advance_tokens(self, max_tokens_count: int) -> int:
        tokens_count: int = 0
        while tokens_count < max_tokens_count:
            if self._remain_tokens_count > 0:
                will_count_tokens = max_tokens_count - tokens_count
                if will_count_tokens > self._remain_tokens_count:
                    tokens_count += self._remain_tokens_count
                    self._remain_tokens_count = 0
                else:
                    tokens_count += will_count_tokens
                    self._remain_tokens_count -= will_count_tokens
                if tokens_count >= max_tokens_count:
                    break
            next_text = self._next_text()
            if next_text is None:
                break
            self._remain_tokens_count += len(self._encoding.encode(next_text))

        return tokens_count

    def _next_text(self) -> str | None:
        next_text: str | None = None
        if self._next_text_buffer is None:
            next_text = next(self._text_iter, None)
        else:
            next_text = self._next_text_buffer
            self._next_text_buffer = None
        return next_text



from pathlib import Path
from xml.etree import ElementTree as ET
from xml.etree.ElementTree import Element

from .zip import Zip


def find_opf_path(zip: Zip) -> Path:
    container_path = Path("META-INF", "container.xml")

    with zip.read(container_path) as f:
        content = f.read()
        root = ET.fromstring(content)
        ns = {"ns": "urn:oasis:names:tc:opendocument:xmlns:container"}
        rootfile = root.find(".//ns:rootfile", ns)

        if rootfile is None:
            rootfile = root.find(".//rootfile")

        if rootfile is None:
            raise ValueError("Cannot find rootfile in container.xml")

        full_path = rootfile.get("full-path")
        if full_path is None:
            raise ValueError("rootfile element has no full-path attribute")

        return Path(full_path)


def strip_namespace(elem: Element) -> None:
    if elem.tag.startswith("{"):
        elem.tag = elem.tag.split("}", 1)[1]

    for child in elem:
        strip_namespace(child)


def extract_namespace(tag: str) -> str | None:
    if tag.startswith("{"):
        parts = tag.split("}", 1)
        if len(parts) == 2:
            return parts[0][1:]
    return None



"""渐进式锁定验证器 - 引导 LLM 逐步收敛到正确结果"""
# pylint: disable=protected-access  # Intentional access to _ValidationContext internals

from xml.etree.ElementTree import Element

from .const import ID_KEY
from .format import _ValidationContext


class ProgressiveLockingValidator:
    """
    渐进式锁定验证器：
    - 验证时收集所有错误（包括已锁定区域）
    - 识别无错误的子树并锁定
    - 只报告未锁定区域的错误
    - 保证收敛：每轮至少锁定一个节点，最多 N 轮完成
    """

    def __init__(self):
        self.locked_ids: set[int] = set()
        self.no_progress_count: int = 0
        self.lock_history: list[set[int]] = []  # 记录每轮锁定的节点

    def validate_with_locking(
        self,
        template_ele: Element,
        validated_ele: Element,
        errors_limit: int,
    ) -> tuple[bool, str | None, set[int]]:
        """
        使用渐进式锁定进行验证

        返回：
        - is_complete: 是否所有节点都已锁定（完成）
        - error_message: 未锁定区域的错误消息（None 表示无错误）
        - newly_locked: 本轮新锁定的节点 ID 集合
        """

        # 1. 执行完整验证（包括已锁定区域）
        context = _ValidationContext()
        context.validate(raw_ele=template_ele, validated_ele=validated_ele)

        # 2. 获取所有错误（以路径为 key）
        all_errors = context._errors

        # 3. 识别可以新锁定的节点
        newly_locked = self._find_lockable_nodes(template_ele, validated_ele, all_errors)

        # 4. 检测卡住情况并解锁
        if not newly_locked and self.locked_ids:
            self.no_progress_count += 1
            if self.no_progress_count >= 3:
                # 卡住了，解锁最近的 2 个节点重试
                self._unlock_recent(count=2)
                self.no_progress_count = 0
        else:
            self.no_progress_count = 0

        # 5. 更新锁定集合
        self.locked_ids.update(newly_locked)
        self.lock_history.append(newly_locked.copy())

        # 6. 过滤错误：只保留未锁定区域的错误
        unlocked_errors = self._filter_unlocked_errors(all_errors)

        # 7. 生成错误消息
        error_message = self._format_errors(unlocked_errors, errors_limit, template_ele)

        # 8. 检查是否完成
        total_nodes = self._count_nodes_with_id(template_ele)
        is_complete = len(self.locked_ids) == total_nodes and error_message is None

        return is_complete, error_message, newly_locked

    def _find_lockable_nodes(
        self, template_ele: Element, validated_ele: Element, errors: dict[tuple[int, ...], list[str]]
    ) -> set[int]:
        """
        找到可以锁定的节点（该节点及所有后代都无错误，且尚未锁定）

        策略：自底向上，优先锁定叶子节点
        """
        lockable = set()

        # 收集所有带 id 的节点，按深度排序（深度优先）
        nodes_with_depth = []
        for elem in template_ele.iter():
            elem_id_str = elem.get(ID_KEY)
            if elem_id_str is not None:
                elem_id = int(elem_id_str)
                if elem_id not in self.locked_ids:
                    depth = self._get_depth(elem, template_ele)
                    nodes_with_depth.append((depth, elem_id, elem))

        # 从最深的节点开始检查
        nodes_with_depth.sort(reverse=True, key=lambda x: x[0])

        for depth, elem_id, elem in nodes_with_depth:
            # 检查该节点的子树是否完全无错误
            if self._subtree_is_error_free(elem, template_ele, errors):
                # 验证在 validated_ele 中也存在对应节点
                validated_node = self._find_by_id(validated_ele, elem_id)
                if validated_node is not None:
                    lockable.add(elem_id)

        return lockable

    def _subtree_is_error_free(
        self, root: Element, template_root: Element, errors: dict[tuple[int, ...], list[str]]
    ) -> bool:
        """检查子树是否完全无错误"""

        # 获取该节点在 template 中的路径
        root_path = self._get_path_to_node(root, template_root)
        if root_path is None:
            return False

        # 检查该路径及其所有后代路径是否有错误
        for error_path in errors.keys():
            # 如果错误路径是 root_path 的后代或等于 root_path
            if self._is_descendant_path(error_path, root_path):
                return False

        return True

    def _get_path_to_node(self, target: Element, root: Element) -> tuple[int, ...] | None:
        """获取从 root 到 target 的路径（以 id 序列表示）"""

        def find_path(current: Element, path: list[int]) -> list[int] | None:
            if current is target:
                return path

            current_id_str = current.get(ID_KEY)
            if current_id_str is not None:
                current_path = path + [int(current_id_str)]
            else:
                current_path = path

            for child in current:
                result = find_path(child, current_path)
                if result is not None:
                    return result

            return None

        path = find_path(root, [])
        return tuple(path) if path is not None else None

    def _is_descendant_path(self, path: tuple[int, ...], ancestor_path: tuple[int, ...]) -> bool:
        """检查 path 是否是 ancestor_path 的后代或相等"""
        if len(path) < len(ancestor_path):
            return False
        return path[: len(ancestor_path)] == ancestor_path

    def _get_depth(self, elem: Element, root: Element) -> int:
        """获取元素的深度"""
        path = self._get_path_to_node(elem, root)
        return len(path) if path else 0

    def _find_by_id(self, root: Element, target_id: int) -> Element | None:
        """在树中查找指定 id 的元素"""
        for elem in root.iter():
            elem_id_str = elem.get(ID_KEY)
            if elem_id_str is not None and int(elem_id_str) == target_id:
                return elem
        return None

    def _filter_unlocked_errors(self, errors: dict[tuple[int, ...], list[str]]) -> dict[tuple[int, ...], list[str]]:
        """过滤错误：只保留路径中包含未锁定节点的错误"""
        unlocked_errors = {}

        for path, error_list in errors.items():
            # 检查路径中是否有未锁定的节点
            has_unlocked = any(node_id not in self.locked_ids for node_id in path)
            if has_unlocked:
                unlocked_errors[path] = error_list

        return unlocked_errors

    def _format_errors(
        self, errors: dict[tuple[int, ...], list[str]], limit: int, template_ele: Element
    ) -> str | None:
        """格式化错误消息（复用现有逻辑）"""
        if not errors:
            return None

        # 使用现有的错误格式化逻辑
        context = _ValidationContext()
        context._errors = errors

        # 构造 _tag_text_dict，从 template_ele 中提取真实的标签信息
        id_to_elem: dict[int, Element] = {}
        for elem in template_ele.iter():
            elem_id_str = elem.get(ID_KEY)
            if elem_id_str is not None:
                elem_id = int(elem_id_str)
                id_to_elem[elem_id] = elem

        # 填充 _tag_text_dict
        for path in errors.keys():
            for node_id in path:
                if node_id not in context._tag_text_dict:
                    elem = id_to_elem.get(node_id)
                    if elem is not None:
                        context._tag_text_dict[node_id] = self._str_tag(elem)
                    else:
                        context._tag_text_dict[node_id] = f'<tag id="{node_id}">'

        return context.errors(limit=limit)

    def _str_tag(self, ele: Element) -> str:
        """生成标签的字符串表示（与 format.py 中的逻辑一致）"""
        ele_id = ele.get(ID_KEY)
        content: str
        if ele_id is not None:
            content = f'<{ele.tag} id="{ele_id}"'
        else:
            content = f"<{ele.tag}"
        if len(ele) > 0:
            content += f"> ... </{ele.tag}>"
        else:
            content += " />"
        return content

    def _count_nodes_with_id(self, root: Element) -> int:
        """统计带有 id 属性的节点数量"""
        count = 0
        for elem in root.iter():
            if elem.get(ID_KEY) is not None:
                count += 1
        return count

    def _unlock_recent(self, count: int):
        """解锁最近锁定的 count 个节点"""
        if not self.lock_history:
            return

        unlocked_count = 0
        # 从最近的历史记录开始解锁
        for i in range(len(self.lock_history) - 1, -1, -1):
            if unlocked_count >= count:
                break

            locked_in_round = self.lock_history[i]
            for node_id in locked_in_round:
                if unlocked_count >= count:
                    break
                if node_id in self.locked_ids:
                    self.locked_ids.remove(node_id)
                    unlocked_count += 1

    def get_progress_summary(self, total_nodes: int) -> str:
        """获取进度摘要"""
        locked_count = len(self.locked_ids)
        percentage = (locked_count / total_nodes * 100) if total_nodes > 0 else 0
        return f"{locked_count}/{total_nodes} nodes locked ({percentage:.1f}%)"



from collections.abc import Generator
from dataclasses import dataclass
from enum import Enum, auto
from io import StringIO


class TagKind(Enum):
    OPENING = auto()
    CLOSING = auto()
    SELF_CLOSING = auto()


@dataclass
class Tag:
    kind: TagKind
    name: str
    proto: str
    attributes: list[tuple[str, str]]

    def __str__(self):
        buffer = StringIO()
        buffer.write("<")
        if self.kind == TagKind.CLOSING:
            buffer.write("/")
        buffer.write(self.name)
        if len(self.attributes) > 0:
            buffer.write(" ")
            for i, (attr_name, attr_value) in enumerate(self.attributes):
                buffer.write(attr_name)
                buffer.write("=")
                buffer.write('"')
                buffer.write(attr_value)
                buffer.write('"')
                if i < len(self.attributes) - 1:
                    buffer.write(" ")
        if self.kind == TagKind.SELF_CLOSING:
            buffer.write("/>")
        else:
            buffer.write(">")
        return buffer.getvalue()

    def find_invalid_name(self) -> str | None:
        for name in self._iter_tag_names():
            if not all(is_valid_value_char(c) for c in name):
                return name
            # https://www.w3schools.com/xml/xml_elements.asp
            # The following logic enforces a subset of XML naming rules:
            # - Names must not be empty.
            # - Names must start with a letter (a-z, A-Z) or an underscore (_).
            if name == "":
                return name
            char = name[0]
            if char == "_":
                continue
            if "a" <= char <= "z" or "A" <= char <= "Z":
                continue
            return name

        return None

    def find_invalid_attr_value(self) -> tuple[str, str] | None:
        for attr_name, attr_value in self.attributes:
            if not all(is_valid_value_char(c) for c in attr_value):
                return attr_name, attr_value
        return None

    def _iter_tag_names(self) -> Generator[str, None, None]:
        yield self.name
        for attr_name, _ in self.attributes:
            yield attr_name


# XML Attribute Values: https://www.w3.org/TR/xml/#NT-AttValue
# URI Syntax: https://www.rfc-editor.org/rfc/rfc3986
# HTML Attributes: https://html.spec.whatwg.org/multipage/syntax.html#attributes-2
_VALID_VALUE_CHARS = frozenset(
    (
        ",",
        ".",
        "/",
        "#",
        "?",
        "&",
        "=",
        ":",
        "%",
        ";",
        " ",
    )
)


# XML Names: https://www.w3.org/TR/xml/#NT-Name
# XML Namespaces: https://www.w3.org/TR/xml-names/#ns-qualnames
# HTML Custom Data Attributes: https://html.spec.whatwg.org/multipage/dom.html#custom-data-attribute
_VALID_NAME_CHARS = frozenset(("-", "_", ":", "."))


def is_valid_value_char(char: str) -> bool:
    if is_valid_name_char(char):
        return True
    if char in _VALID_VALUE_CHARS:
        return True
    return False


def is_valid_name_char(char: str) -> bool:
    if char in _VALID_NAME_CHARS:
        return True

    # https://www.w3.org/TR/xml/#NT-Name
    if "a" <= char <= "z":
        return True
    if "A" <= char <= "Z":
        return True
    if "0" <= char <= "9":
        return True
    return False



from xml.etree.ElementTree import Element

# 运算符映射表
_OPERATOR_MAP = {
    "→": r"\rightarrow",
    "←": r"\leftarrow",
    "↔": r"\leftrightarrow",
    "×": r"\times",
    "·": r"\cdot",
    "÷": r"\div",
    "±": r"\pm",
    "∓": r"\mp",
    "≤": r"\leq",
    "≥": r"\geq",
    "≠": r"\neq",
    "≈": r"\approx",
    "∞": r"\infty",
    "∫": r"\int",
    "∑": r"\sum",
    "∏": r"\prod",
    "√": r"\sqrt",
    "∂": r"\partial",
    "∇": r"\nabla",
    "∈": r"\in",
    "∉": r"\notin",
    "⊂": r"\subset",
    "⊃": r"\supset",
    "⊆": r"\subseteq",
    "⊇": r"\supseteq",
    "∪": r"\cup",
    "∩": r"\cap",
    "∅": r"\emptyset",
    "∀": r"\forall",
    "∃": r"\exists",
    "¬": r"\neg",
    "∧": r"\land",
    "∨": r"\lor",
    "α": r"\alpha",
    "β": r"\beta",
    "γ": r"\gamma",
    "δ": r"\delta",
    "ε": r"\epsilon",
    "θ": r"\theta",
    "λ": r"\lambda",
    "μ": r"\mu",
    "π": r"\pi",
    "σ": r"\sigma",
    "φ": r"\phi",
    "ω": r"\omega",
    "Δ": r"\Delta",
    "Σ": r"\Sigma",
    "Ω": r"\Omega",
}


def xml_to_latex(element: Element) -> str:
    tag = element.tag

    # 根据元素类型进行转换
    if tag == "math":
        # 根元素，只处理子元素
        return "".join(xml_to_latex(child) for child in element)

    elif tag == "mrow":
        # 分组元素，递归处理所有子元素
        return "".join(xml_to_latex(child) for child in element)

    elif tag == "mi":
        # 标识符（变量名）
        text = element.text or ""
        # 多字符标识符用 \mathrm
        if len(text) > 1:
            return f"\\mathrm{{{text}}}"
        return text

    elif tag == "mn":
        # 数字
        return element.text or ""

    elif tag == "mo":
        # 运算符
        text = (element.text or "").strip()
        return _OPERATOR_MAP.get(text, text)

    elif tag == "mfrac":
        # 分数
        children = list(element)
        if len(children) >= 2:
            numerator = xml_to_latex(children[0])
            denominator = xml_to_latex(children[1])
            return f"\\frac{{{numerator}}}{{{denominator}}}"
        return ""

    elif tag == "msub":
        # 下标
        children = list(element)
        if len(children) >= 2:
            base = xml_to_latex(children[0])
            subscript = xml_to_latex(children[1])
            return f"{base}_{{{subscript}}}"
        return ""

    elif tag == "msup":
        # 上标
        children = list(element)
        if len(children) >= 2:
            base = xml_to_latex(children[0])
            superscript = xml_to_latex(children[1])
            return f"{base}^{{{superscript}}}"
        return ""

    elif tag == "msubsup":
        # 同时有上下标
        children = list(element)
        if len(children) >= 3:
            base = xml_to_latex(children[0])
            subscript = xml_to_latex(children[1])
            superscript = xml_to_latex(children[2])
            return f"{base}_{{{subscript}}}^{{{superscript}}}"
        return ""

    elif tag == "msqrt":
        # 平方根
        content = "".join(xml_to_latex(child) for child in element)
        return f"\\sqrt{{{content}}}"

    elif tag == "mroot":
        # n次根
        children = list(element)
        if len(children) >= 2:
            base = xml_to_latex(children[0])
            index = xml_to_latex(children[1])
            return f"\\sqrt[{index}]{{{base}}}"
        return ""

    elif tag == "munder":
        # 下方符号
        children = list(element)
        if len(children) >= 2:
            base = xml_to_latex(children[0])
            under = xml_to_latex(children[1])
            return f"\\underset{{{under}}}{{{base}}}"
        return ""

    elif tag == "mover":
        # 上方符号
        children = list(element)
        if len(children) >= 2:
            base = xml_to_latex(children[0])
            over = xml_to_latex(children[1])
            return f"\\overset{{{over}}}{{{base}}}"
        return ""

    elif tag == "munderover":
        # 上下方符号
        children = list(element)
        if len(children) >= 3:
            base = xml_to_latex(children[0])
            under = xml_to_latex(children[1])
            over = xml_to_latex(children[2])
            # 特殊处理求和、积分等
            base_str = base.strip()
            if base_str in (r"\sum", r"\int", r"\prod"):
                return f"{base}_{{{under}}}^{{{over}}}"
            return f"\\overset{{{over}}}{{\\underset{{{under}}}{{{base}}}}}"
        return ""

    elif tag == "mtext":
        # 文本
        text = element.text or ""
        return f"\\text{{{text}}}"

    elif tag == "mspace":
        # 空格
        return r"\,"

    elif tag == "mtable":
        # 表格/矩阵
        rows = [xml_to_latex(child) for child in element if child.tag.endswith("mtr")]
        return f"\\begin{{array}}{{{rows[0].count('&') + 1}}}\n" + "\\\\\n".join(rows) + "\n\\end{array}"

    elif tag == "mtr":
        # 表格行
        cells = [xml_to_latex(child) for child in element if child.tag.endswith("mtd")]
        return " & ".join(cells)

    elif tag == "mtd":
        # 表格单元格
        return "".join(xml_to_latex(child) for child in element)

    else:
        # 未知元素，递归处理子元素
        return "".join(xml_to_latex(child) for child in element)



from xml.etree.ElementTree import Element

from ..utils import normalize_whitespace
from ..xml import plain_text
from .const import DATA_ORIGIN_LEN_KEY, ID_KEY
from .format import format
from .text_segment import TextSegment, combine_text_segments


class XMLFill:
    def __init__(self, text_segments: list[TextSegment]) -> None:
        self._request_element = Element("xml")
        self._text_segments: dict[tuple[int, ...], list[TextSegment]] = {}  # generated id stack -> text segments

        raw2generated: dict[int, Element] = {}
        raw2generated_ids: dict[int, int] = {}

        for combined_element, sub_raw2generated in combine_text_segments(text_segments):
            unwrapped_parent_ids: set[int] = set()
            sub_element, parents = self._unwrap_parents(combined_element)
            self._request_element.append(sub_element)
            for parent in parents:
                unwrapped_parent_ids.add(id(parent))

            for raw_id, generated_element in sub_raw2generated.items():
                if raw_id in unwrapped_parent_ids:
                    continue
                if id(generated_element) in unwrapped_parent_ids:
                    continue
                generated_id = len(raw2generated)
                raw2generated[raw_id] = generated_element
                raw2generated_ids[raw_id] = generated_id

                generated_text = normalize_whitespace(
                    text=plain_text(generated_element),
                )
                generated_element.attrib = {
                    ID_KEY: str(generated_id),
                    DATA_ORIGIN_LEN_KEY: str(len(generated_text)),
                }

        for text_segment in text_segments:
            generated_id_stack: list[int] = []
            for parent in text_segment.parent_stack:
                generated_id = raw2generated_ids.get(id(parent), None)
                if generated_id is not None:
                    generated_id_stack.append(generated_id)
            generated_key = tuple(generated_id_stack)
            text_segments_stack = self._text_segments.get(generated_key, None)
            if text_segments_stack is None:
                text_segments_stack = []
                self._text_segments[generated_key] = text_segments_stack
            text_segments_stack.append(text_segment)

        for text_segments_stack in self._text_segments.values():
            text_segments_stack.reverse()  # for use call .pop()

    def _unwrap_parents(self, element: Element):
        parents: list[Element] = []
        while True:
            if len(element) != 1:
                break
            child = element[0]
            if not element.text:
                break
            if not child.tail:
                break
            parents.append(element)
            element = child
            element.tail = None
        return element, parents

    @property
    def request_element(self) -> Element:
        return self._request_element

    def submit_response_text(self, text: str, errors_limit: int) -> Element:
        submitted_element = format(
            template_ele=self._request_element,
            validated_text=text,
            errors_limit=errors_limit,
        )
        self._fill_submitted_texts(
            generated_ids_stack=[],
            element=submitted_element,
        )
        return submitted_element

    def _fill_submitted_texts(self, generated_ids_stack: list[int], element: Element):
        current_stack = generated_ids_stack
        generated_id = self._generated_id(element)
        if generated_id >= 0:
            current_stack = generated_ids_stack + [generated_id]

        generated_key = tuple(current_stack)
        text_segments_stack = self._text_segments.get(generated_key, None)
        text = self._normalize_text(element.text)

        if text_segments_stack and text is not None:
            text_segment = text_segments_stack.pop()
            text_segment.text = text

        for child_element in element:
            self._fill_submitted_texts(
                generated_ids_stack=current_stack,
                element=child_element,
            )
            tail = self._normalize_text(child_element.tail)
            if text_segments_stack and tail is not None:
                text_segment = text_segments_stack.pop()
                text_segment.text = tail

    def _generated_id(self, element: Element) -> int:
        str_id = element.get(ID_KEY, None)
        if str_id is None:
            return -1
        try:
            return int(str_id)
        except ValueError:
            return -1

    def _normalize_text(self, text: str | None) -> str | None:
        if text is None:
            return None
        text = normalize_whitespace(text)
        if not text.strip():
            return None
        return text



import httpx
import openai
import requests


def is_retry_error(err: Exception) -> bool:
    if _is_openai_retry_error(err):
        return True
    if _is_httpx_retry_error(err):
        return True
    if _is_request_retry_error(err):
        return True
    return False


# https://help.openai.com/en/articles/6897213-openai-library-error-types-guidance
def _is_openai_retry_error(err: Exception) -> bool:
    if isinstance(err, openai.Timeout):
        return True
    if isinstance(err, openai.APIConnectionError):
        return True
    if isinstance(err, openai.InternalServerError):
        return err.status_code in (502, 503, 504)
    return False


# https://www.python-httpx.org/exceptions/
def _is_httpx_retry_error(err: Exception) -> bool:
    if isinstance(err, httpx.RemoteProtocolError):
        return True
    if isinstance(err, httpx.StreamError):
        return True
    if isinstance(err, httpx.TimeoutException):
        return True
    if isinstance(err, httpx.NetworkError):
        return True
    if isinstance(err, httpx.ProtocolError):
        return True
    return False


# https://requests.readthedocs.io/en/latest/api/#exceptions
def _is_request_retry_error(err: Exception) -> bool:
    if isinstance(err, requests.ConnectionError):
        return True
    if isinstance(err, requests.ConnectTimeout):
        return True
    if isinstance(err, requests.ReadTimeout):
        return True
    if isinstance(err, requests.Timeout):
        return True
    return False



from collections.abc import Generator, Iterable
from xml.etree.ElementTree import Element

from ..xml import clone_element
from .parser import parse_tags
from .tag import Tag, TagKind
from .transform import tag_to_element


# why implement XML decoding?
# https://github.com/oomol-lab/pdf-craft/issues/149
def decode_friendly(chars: Iterable[str], tags: Iterable[str] | str = ()) -> Generator[Element, None, None]:
    if isinstance(tags, str):
        tags = set((tags,))
    else:
        tags = set(tags)

    for element in _collect_elements(chars):
        if element.tag in tags or len(tags) == 0:
            yield clone_element(element)


def _collect_elements(chars: Iterable[str]) -> Generator[Element, None, None]:
    opening_stack: list[Element] = []
    last_closed_element: Element | None = None

    for cell in parse_tags(chars):
        if isinstance(cell, Tag):
            tag: Tag = cell
            element = tag_to_element(tag)
            if tag.kind == TagKind.CLOSING:
                popped = _pop_element(tag.name, opening_stack)
                if popped is not None:
                    yield popped
                    last_closed_element = popped
                elif last_closed_element is not None:
                    _append_to_tail(last_closed_element, tag.proto)
            else:
                if opening_stack:
                    opening_stack[-1].append(element)
                if tag.kind == TagKind.SELF_CLOSING:
                    yield element
                    last_closed_element = element
                elif tag.kind == TagKind.OPENING:
                    opening_stack.append(element)
                    last_closed_element = None

        elif last_closed_element is not None:
            _append_to_tail(last_closed_element, cell)

        elif opening_stack:
            opening_stack[-1].text = cell


def _append_to_tail(element: Element, text: str) -> None:
    if element.tail:
        element.tail += text
    else:
        element.tail = text


def _pop_element(tag_name: str, opening_stack: list[Element]) -> Element | None:
    index = -1
    for i in range(len(opening_stack) - 1, -1, -1):
        opening_element = opening_stack[i]
        if tag_name == opening_element.tag:
            index = i
            break
    if index == -1:
        return None

    popped: Element | None = None
    for _ in range(len(opening_stack) - index):
        popped = opening_stack.pop()
    return popped



from collections.abc import Generator, Iterable
from dataclasses import dataclass
from enum import Enum, auto
from typing import Self
from xml.etree.ElementTree import Element

from .utils import expand_left_element_texts, expand_right_element_texts, normalize_text_in_element

# HTML inline-level elements
# Reference: https://developer.mozilla.org/en-US/docs/Web/HTML/Inline_elements
# Reference: https://developer.mozilla.org/en-US/docs/Glossary/Inline-level_content
_HTML_INLINE_TAGS = frozenset(
    [
        # Inline text semantics
        "a",
        "abbr",
        "b",
        "bdi",
        "bdo",
        "br",
        "cite",
        "code",
        "data",
        "dfn",
        "em",
        "i",
        "kbd",
        "mark",
        "q",
        "rp",
        "rt",
        "ruby",
        "s",
        "samp",
        "small",
        "span",
        "strong",
        "sub",
        "sup",
        "time",
        "u",
        "var",
        "wbr",
        # Image and multimedia
        "img",
        "svg",
        "canvas",
        "audio",
        "video",
        "map",
        "area",
        # Form elements
        "input",
        "button",
        "select",
        "textarea",
        "label",
        "output",
        "progress",
        "meter",
        # Embedded content
        "iframe",
        "embed",
        "object",
        # Other inline elements
        "script",
        "del",
        "ins",
        "slot",
    ]
)


class TextPosition(Enum):
    TEXT = auto()
    TAIL = auto()


@dataclass
class TextSegment:
    text: str
    index: int  # *.text is 0, the first *.tail is 1, and so on
    parent_stack: list[Element]
    left_common_depth: int
    right_common_depth: int
    block_depth: int
    position: TextPosition

    @property
    def root(self) -> Element:
        return self.parent_stack[0]

    @property
    def block_parent(self) -> Element:
        return self.parent_stack[self.block_depth - 1]

    @property
    def xml_text(self) -> str:
        return "".join(_expand_xml_texts(self))

    def strip_block_parents(self) -> Self:
        self.parent_stack = self.parent_stack[self.block_depth - 1 :]
        self.block_depth = 1
        return self

    def clone(self) -> "TextSegment":
        return TextSegment(
            text=self.text,
            index=self.index,
            parent_stack=list(self.parent_stack),
            left_common_depth=self.left_common_depth,
            right_common_depth=self.right_common_depth,
            block_depth=self.block_depth,
            position=self.position,
        )


def _expand_xml_texts(segment: TextSegment):
    for i in range(segment.left_common_depth, len(segment.parent_stack)):
        yield from expand_left_element_texts(segment.parent_stack[i])
    yield segment.text
    for i in range(len(segment.parent_stack) - 1, segment.right_common_depth - 1, -1):
        yield from expand_right_element_texts(segment.parent_stack[i])


def incision_between(segment1: TextSegment, segment2: TextSegment) -> tuple[int, int]:
    return (
        _incision_of(segment1, segment1.right_common_depth),
        _incision_of(segment2, segment2.left_common_depth),
    )


def _incision_of(segment: TextSegment, common_depth: int) -> int:
    block_diff: int = 0
    inline_diff: int = 0
    if common_depth >= segment.block_depth:
        inline_diff = len(segment.parent_stack) - common_depth
    else:
        block_diff = segment.block_depth - common_depth
        inline_diff = len(segment.parent_stack) - segment.block_depth
    return block_diff * 3 + inline_diff  # 数字越大越容易被拆分


def search_text_segments(root: Element) -> Generator[TextSegment, None, None]:
    generator = _search_text_segments([], root)
    text_segment = next(generator, None)
    if text_segment is None:
        return

    while True:
        next_text_segment = next(generator, None)
        if next_text_segment is None:
            break
        common_depth = _common_depth(
            stack1=text_segment.parent_stack,
            stack2=next_text_segment.parent_stack,
        )
        text_segment.right_common_depth = common_depth
        yield text_segment
        text_segment = next_text_segment
        text_segment.left_common_depth = common_depth

    yield text_segment


def _search_text_segments(stack: list[Element], element: Element) -> Generator[TextSegment, None, None]:
    text = normalize_text_in_element(element.text)
    next_stack = stack + [element]
    next_block_depth = _find_block_depth(next_stack)

    if text is not None:
        yield TextSegment(
            text=text,
            index=0,
            parent_stack=next_stack,
            left_common_depth=0,
            right_common_depth=0,
            block_depth=next_block_depth,
            position=TextPosition.TEXT,
        )
    for i, child_element in enumerate(element):
        yield from _search_text_segments(next_stack, child_element)
        child_tail = normalize_text_in_element(child_element.tail)
        if child_tail is not None:
            yield TextSegment(
                text=child_tail,
                index=i + 1,
                parent_stack=next_stack,
                left_common_depth=0,
                right_common_depth=0,
                block_depth=next_block_depth,
                position=TextPosition.TAIL,
            )


def _find_block_depth(parent_stack: list[Element]) -> int:
    index: int = 0
    for i in range(len(parent_stack) - 1, -1, -1):
        checked_tag = parent_stack[i].tag.lower()
        if checked_tag not in _HTML_INLINE_TAGS:
            index = i
            break
    return index + 1  # depth is a count not index


def combine_text_segments(segments: Iterable[TextSegment]) -> Generator[tuple[Element, dict[int, Element]], None, None]:
    stack: list[tuple[Element, Element]] = []  # (raw, generated)
    raw2generated: dict[int, Element] = {}
    last_popped: Element | None = None

    for segment in segments:
        common_depth = _common_depth(
            stack1=(raw for raw, _ in stack),
            stack2=segment.parent_stack,
        )
        if stack and common_depth == 0:
            yield stack[0][1], raw2generated
            stack = []
            raw2generated = {}
            last_popped = None

        while len(stack) > common_depth:
            last_popped = stack.pop()[1]
        while len(stack) < len(segment.parent_stack):
            last_popped = None
            index = len(stack)
            raw = segment.parent_stack[index]
            generated = Element(raw.tag, raw.attrib)
            if stack:
                _, generated_parent = stack[-1]
                generated_parent.append(generated)
            stack.append((raw, generated))
            raw2generated[id(raw)] = generated

        if last_popped is None:
            if stack:
                stack[-1][1].text = _append_element_text(
                    text=stack[-1][1].text,
                    appended=segment.text,
                )
        else:
            last_popped.tail = _append_element_text(
                text=last_popped.tail,
                appended=segment.text,
            )
    if stack:
        yield stack[0][1], raw2generated


def _common_depth(stack1: Iterable[Element], stack2: Iterable[Element]) -> int:
    common_depth: int = 0
    for parent1, parent2 in zip(stack1, stack2):
        if id(parent1) != id(parent2):
            break
        common_depth += 1
    return common_depth


def _append_element_text(text: str | None, appended: str) -> str:
    if text is None:
        return appended
    else:
        return text + appended



from xml.etree.ElementTree import Element

from ..utils import normalize_whitespace
from ..xml import decode_friendly
from .const import ID_KEY


def format(template_ele: Element, validated_text: str, errors_limit: int) -> Element:
    context = _ValidationContext()
    validated_ele = _extract_xml_element(validated_text)
    context.validate(raw_ele=template_ele, validated_ele=validated_ele)
    error_message = context.errors(limit=errors_limit)
    if error_message:
        raise ValidationError(message=error_message, validated_ele=validated_ele)
    return validated_ele


class ValidationError(Exception):
    def __init__(self, message: str, validated_ele: Element | None = None) -> None:
        super().__init__(message)
        self.validated_ele = validated_ele


def _extract_xml_element(text: str) -> Element:
    first_xml_element: Element | None = None
    all_xml_elements: int = 0

    for xml_element in decode_friendly(text, tags="xml"):
        if first_xml_element is None:
            first_xml_element = xml_element
        all_xml_elements += 1

    if first_xml_element is None:
        raise ValidationError(
            "No complete <xml>...</xml> block found. Please ensure you have properly closed the XML with </xml> tag."
        )
    if all_xml_elements > 1:
        raise ValidationError(
            f"Found {all_xml_elements} <xml>...</xml> blocks. "
            "Please return only one XML block without any examples or explanations."
        )
    return first_xml_element


class _ValidationContext:
    def __init__(self) -> None:
        self._tag_text_dict: dict[int, str] = {}
        self._errors: dict[tuple[int, ...], list[str]] = {}

    def validate(self, raw_ele: Element, validated_ele: Element):
        self._validate_ele(ids_path=[], raw_ele=raw_ele, validated_ele=validated_ele)

    def errors(self, limit: int) -> str | None:
        if not self._errors:
            return

        keys = list(self._errors.keys())
        keys.sort(key=lambda k: (len(k), k))  # AI 矫正应该先浅后深
        keys = keys[:limit]
        max_len_key = max((len(key) for key in keys), default=0)

        for i in range(len(keys)):
            key = keys[i]
            if len(key) < max_len_key:
                key_list = list(key)
                while len(key_list) < max_len_key:
                    key_list.append(-1)
                keys[i] = tuple(key_list)

        content: list[str] = []
        total_errors = sum(len(messages) for messages in self._errors.values())
        remain_errors = total_errors

        for key in sorted(keys):  # 改成深度优先排序，看起来关联度更好
            raw_key = tuple(k for k in key if k >= 0)
            indent: str = f"{'  ' * len(raw_key)}"
            errors_list = self._errors[raw_key]
            parent_text: str

            if len(raw_key) > 0:
                parent_text = self._tag_text_dict[raw_key[-1]]
            else:
                parent_text = "the root tag"

            if len(errors_list) == 1:
                error = errors_list[0]
                content.append(f"{indent}- errors in {parent_text}: {error}.")
            else:
                content.append(f"{indent}- errors in {parent_text}:")
                for error in errors_list:
                    content.append(f"{indent}  - {error}.")
            remain_errors -= len(errors_list)

        content.insert(0, f"Found {total_errors} error(s) in your response XML structure.")
        if remain_errors > 0:
            content.append(f"\n... and {remain_errors} more error(s).")

        return "\n".join(content)

    def _validate_ele(self, ids_path: list[int], raw_ele: Element, validated_ele: Element):
        raw_id_map = self._build_id_map(raw_ele)
        validated_id_map = self._build_id_map(validated_ele)
        lost_ids: list[int] = []
        extra_ids: list[int] = []

        for id, sub_raw in raw_id_map.items():
            sub_validated = validated_id_map.get(id, None)
            if sub_validated is None:
                lost_ids.append(id)
            else:
                self._validate_id_ele(
                    id=id,
                    ids_path=ids_path,
                    raw_ele=sub_raw,
                    validated_ele=sub_validated,
                )

        for id in validated_id_map.keys():
            if id not in raw_id_map:
                extra_ids.append(id)

        if lost_ids or extra_ids:
            messages: list[str] = []
            lost_ids.sort()
            extra_ids.sort()

            if lost_ids:
                tags = [self._str_tag(raw_id_map[id]) for id in lost_ids]
                # Provide context from source XML
                context_info = self._get_source_context(raw_ele, lost_ids)
                messages.append(f"lost sub-tags {' '.join(tags)}")
                if context_info:
                    messages.append(f"Source structure was: {context_info}")

            if extra_ids:
                tags = [self._str_tag(validated_id_map[id]) for id in extra_ids]
                messages.append(f"extra sub-tags {' '.join(tags)}")

            if messages:
                self._add_error(
                    ids_path=ids_path,
                    message="find " + " and ".join(messages),
                )
        else:
            raw_element_empty = not self._has_text_content(raw_ele)
            validated_ele_empty = not self._has_text_content(validated_ele)

            if raw_element_empty and not validated_ele_empty:
                self._add_error(
                    ids_path=ids_path,
                    message="shouldn't have text content",
                )
            elif not raw_element_empty and validated_ele_empty:
                self._add_error(
                    ids_path=ids_path,
                    message="text content is missing",
                )

    def _validate_id_ele(self, ids_path: list[int], id: int, raw_ele: Element, validated_ele: Element):
        if raw_ele.tag == validated_ele.tag:
            self._tag_text_dict[id] = self._str_tag(raw_ele)
            raw_has_text = self._has_direct_text(raw_ele.text)
            validated_has_text = self._has_direct_text(validated_ele.text)

            if raw_has_text and not validated_has_text:
                self._add_error(
                    ids_path=ids_path + [id],
                    message="missing text content before child elements",
                )
            elif not raw_has_text and validated_has_text:
                self._add_error(
                    ids_path=ids_path + [id],
                    message="shouldn't have text content before child elements",
                )
            raw_has_tail = self._has_direct_text(raw_ele.tail)
            validated_has_tail = self._has_direct_text(validated_ele.tail)

            if raw_has_tail and not validated_has_tail:
                self._add_error(
                    ids_path=ids_path + [id],
                    message="missing text content after the element",
                )
            elif not raw_has_tail and validated_has_tail:
                self._add_error(
                    ids_path=ids_path + [id],
                    message="shouldn't have text content after the element",
                )

            self._validate_ele(
                ids_path=ids_path + [id],
                raw_ele=raw_ele,
                validated_ele=validated_ele,
            )
        else:
            self._add_error(
                ids_path=ids_path,
                message=f'got <{validated_ele.tag} id="{id}">',
            )

    def _add_error(self, ids_path: list[int], message: str):
        key = tuple(ids_path)
        if key not in self._errors:
            self._errors[key] = []
        self._errors[key].append(message)

    def _build_id_map(self, ele: Element):
        id_map: dict[int, Element] = {}
        for child_ele in ele:
            id_text = child_ele.get(ID_KEY, None)
            if id_text is not None:
                id = int(id_text)
                if id < 0:
                    raise ValueError(f"Invalid id {id} found. IDs must be non-negative integers.")
                if id_text is not None:
                    id_map[id] = child_ele
        return id_map

    def _has_text_content(self, ele: Element) -> bool:
        text = "".join(self._plain_text(ele))
        text = normalize_whitespace(text)
        text = text.strip()
        return len(text) > 0

    def _has_direct_text(self, text: str | None) -> bool:
        if text is None:
            return False
        normalized = normalize_whitespace(text).strip()
        return len(normalized) > 0

    def _plain_text(self, ele: Element):
        if ele.text:
            yield ele.text
        for child in ele:
            if child.get(ID_KEY, None) is not None:
                yield from self._plain_text(child)
            if child.tail:
                yield child.tail

    def _str_tag(self, ele: Element) -> str:
        ele_id = ele.get(ID_KEY)
        content: str
        if ele_id is not None:
            content = f'<{ele.tag} id="{ele_id}"'
        else:
            content = f"<{ele.tag}"
        if len(ele) > 0:
            content += f"> ... </{ele.tag}>"
        else:
            content += " />"
        return content

    def _get_source_context(self, parent: Element, lost_ids: list[int]) -> str:
        """Generate context showing where lost tags appeared in source XML."""
        if not lost_ids:
            return ""

        # Build a simple representation of the source structure
        children_with_ids = []
        for child in parent:
            child_id_str = child.get(ID_KEY)
            if child_id_str is not None:
                child_id = int(child_id_str)
                is_lost = child_id in lost_ids
                tag_str = f'<{child.tag} id="{child_id}">'

                # Show text before/inside/after
                parts = []
                if child.text and child.text.strip():
                    preview = child.text.strip()[:20]
                    if is_lost:
                        parts.append(f'[{preview}...]')
                    else:
                        parts.append(f'{preview}...')

                if is_lost:
                    children_with_ids.append(f'{tag_str}*MISSING*')
                else:
                    children_with_ids.append(tag_str)

        if children_with_ids:
            return f"[{' '.join(children_with_ids)}]"
        return ""



from xml.etree.ElementTree import Element

from ..utils import normalize_whitespace
from .const import DATA_ORIGIN_LEN_KEY, ID_KEY


def normalize_text_in_element(text: str | None) -> str | None:
    if text is None:
        return None
    text = normalize_whitespace(text)
    if not text.strip():
        return None
    return text


def expand_left_element_texts(element: Element):
    yield "<"
    yield element.tag
    yield " "
    yield ID_KEY
    yield '="99" '
    yield DATA_ORIGIN_LEN_KEY
    yield '="999">'


def expand_right_element_texts(element: Element):
    yield "</"
    yield element.tag
    yield ">"



class Increaser:
    def __init__(self, value_range: tuple[float, float] | None):
        self._value_range: tuple[float, float] | None = value_range
        self._current: float | None = value_range[0] if value_range is not None else None

    @property
    def current(self) -> float | None:
        return self._current

    def increase(self):
        if self._value_range is not None and self._current is not None:
            _, end_value = self._value_range
            self._current = self._current + 0.5 * (end_value - self._current)


class Increasable:
    def __init__(self, param: float | tuple[float, float] | None):
        self._value_range: tuple[float, float] | None = None

        if isinstance(param, int):
            param = float(param)
        if isinstance(param, float):
            param = (param, param)
        if isinstance(param, tuple):
            if len(param) != 2:
                raise ValueError(f"Expected a tuple of length 2, got {len(param)}")
            begin, end = param
            if isinstance(begin, int):
                begin = float(begin)
            if isinstance(end, int):
                end = float(end)
            self._value_range = (begin, end)

    def context(self) -> Increaser:
        return Increaser(self._value_range)



from collections.abc import Generator, Iterable
from typing import Generic, TypeVar

T = TypeVar("T")


class IterSync(Generic[T]):
    def __init__(self) -> None:
        super().__init__()
        self._queue: list[T] = []

    @property
    def tail(self) -> T | None:
        if not self._queue:
            return None
        return self._queue[-1]

    def take(self) -> T:
        return self._queue.pop()

    def iter(self, elements: Iterable[T]) -> Generator[T, None, None]:
        for element in elements:
            self._queue.insert(0, element)
            yield element



from collections.abc import Generator
from xml.etree.ElementTree import Element


def find_first(element: Element, tag: str) -> Element | None:
    if element.tag == tag:
        return element
    for child in element:
        result = find_first(child, tag)
        if result is not None:
            return result
    return None


def iter_with_stack(element: Element) -> Generator[tuple[list[Element], Element], None, None]:
    """先序遍历：yield parent_path, element"""
    stack: list[list[Element]] = [[element]]
    while stack:
        current_path = stack.pop()
        current = current_path[-1]
        yield current_path[:-1], current

        if len(current) == 0:
            continue

        for child in reversed(list(current)):
            child_path = list(current_path)
            child_path.append(child)
            stack.append(child_path)


def clone_element(element: Element) -> Element:
    new_element = Element(element.tag, element.attrib)
    new_element.text = element.text
    for child in element:
        new_child = clone_element(child)
        new_child.tail = child.tail
        new_element.append(new_child)
    return new_element


def plain_text(element: Element) -> str:
    return "".join(_iter_text_in(element))


def _iter_text_in(element: Element) -> Generator[str, None, None]:
    if element.text:
        yield element.text
    for child in element:
        yield from _iter_text_in(child)
        if child.tail:
            yield child.tail



import zipfile
from pathlib import Path
from typing import IO

_BUFFER_SIZE = 8192  # 8KB


class Zip:
    def __init__(self, source_path: Path, target_path: Path) -> None:
        source_zip: zipfile.ZipFile | None = None
        target_zip: zipfile.ZipFile | None = None
        try:
            source_zip = zipfile.ZipFile(source_path, "r")
            target_zip = zipfile.ZipFile(target_path, "w", zipfile.ZIP_DEFLATED)
        except Exception:
            if source_zip:
                source_zip.close()
            if target_zip:
                target_zip.close()
            raise
        self._source_zip: zipfile.ZipFile = source_zip
        self._target_zip: zipfile.ZipFile = target_zip
        self._processed_files: set[Path] = set()

    def __enter__(self):
        return self

    def __exit__(self, _exc_type, _exc_val, _exc_tb):
        try:
            if _exc_type is None:
                all_files = self._source_zip.namelist()
                for file_path in all_files:
                    if file_path.endswith("/"):
                        continue
                    if Path(file_path) not in self._processed_files:
                        self.migrate(Path(file_path))
        finally:
            self._target_zip.close()
            self._source_zip.close()

        return False

    def list_files(self, prefix_path: Path | None = None) -> list[Path]:
        all_files = self._source_zip.namelist()
        if prefix_path is None:
            return [Path(f) for f in all_files]
        prefix = str(prefix_path)
        if not prefix.endswith("/"):
            prefix += "/"
        return [Path(f) for f in all_files if f.startswith(prefix)]

    def migrate(self, path: Path):
        with self.read(path) as source_file:
            with self._target_zip.open(str(path), "w") as target_file:
                while True:
                    chunk = source_file.read(_BUFFER_SIZE)
                    if not chunk:
                        break
                    target_file.write(chunk)
        self._processed_files.add(path)

    def read(self, path: Path) -> IO[bytes]:
        return self._source_zip.open(str(path), "r")

    def replace(self, path: Path) -> IO[bytes]:
        self._processed_files.add(path)
        return self._target_zip.open(str(path), "w")



from collections.abc import Callable
from html import escape as escape_html
from io import StringIO
from xml.etree.ElementTree import Element

from .parser import parse_tags
from .tag import Tag, TagKind
from .transform import element_to_tag


# why implement XML encoding?
# https://github.com/oomol-lab/pdf-craft/issues/149
def encode_friendly(element: Element, indent: int = 2) -> str:
    buffer = StringIO()
    _encode_element(
        buffer=buffer,
        element=element,
        indent=indent,
        depth=0,
        escape=_escape_text,
    )
    return buffer.getvalue()


def _escape_text(text: str) -> str:
    buffer = StringIO()
    for cell in parse_tags(text):
        if isinstance(cell, Tag):
            cell = escape_html(str(cell))
        buffer.write(cell)
    return buffer.getvalue()


_TINY_TEXT_LEN = 35


def _encode_element(
    buffer: StringIO,
    element: Element,
    indent: int,
    depth: int,
    escape: Callable[[str], str],
) -> None:
    _write_indent(buffer, indent, depth)
    if len(element) == 0 and not element.text:
        tag = element_to_tag(element, TagKind.SELF_CLOSING)
        buffer.write(str(tag))
    else:
        text = (element.text or "").strip()
        opening_tag = element_to_tag(element, TagKind.OPENING)
        closing_tag = element_to_tag(element, TagKind.CLOSING)
        buffer.write(str(opening_tag))
        is_one_line = len(text) <= _TINY_TEXT_LEN and len(element) == 0 and "\n" not in text
        if text:
            if not is_one_line:
                buffer.write("\n")
                _write_indent(buffer, indent, depth + 1)
            buffer.write(escape(text))

        for child in element:
            buffer.write("\n")
            _encode_element(
                buffer=buffer,
                element=child,
                indent=indent,
                depth=depth + 1,
                escape=escape,
            )
            child_tail = (child.tail or "").strip()
            if child_tail:
                buffer.write("\n")
                _write_indent(buffer, indent, depth + 1)
                buffer.write(escape(child_tail))

        if not is_one_line:
            buffer.write("\n")
            _write_indent(buffer, indent, depth)

        buffer.write(str(closing_tag))


def _write_indent(buffer: StringIO, indent: int, depth: int) -> None:
    for _ in range(indent * depth):
        buffer.write(" ")



from typing import Generic, Protocol, Self, TypeVar, runtime_checkable

S = TypeVar("S", covariant=True)
T = TypeVar("T")
ST = TypeVar("ST", bound="Segment")


@runtime_checkable
class Segment(Protocol, Generic[S]):
    @property
    def tokens(self) -> int: ...

    @property
    def payload(self) -> S: ...

    def truncate_after_head(self, remain_tokens: int) -> Self: ...
    def truncate_before_tail(self, remain_tokens: int) -> Self: ...



from collections.abc import Generator, Iterable
from typing import TypeVar
from xml.etree.ElementTree import Element

from ..iter_sync import IterSync
from ..llm import LLM, Message, MessageRole
from ..xml import encode_friendly
from .fill import XMLFill
from .format import ValidationError, _extract_xml_element
from .group import XMLGroupContext
from .progressive_locking import ProgressiveLockingValidator
from .text_segment import TextSegment

T = TypeVar("T")


class XMLTranslator:
    def __init__(
        self,
        llm: LLM,
        group_context: XMLGroupContext,
        target_language: str,
        user_prompt: str | None,
        ignore_translated_error: bool,
        max_retries: int,
        max_fill_displaying_errors: int,
    ) -> None:
        self._llm: LLM = llm
        self._group_context: XMLGroupContext = group_context
        self._target_language: str = target_language
        self._user_prompt: str | None = user_prompt
        self._ignore_translated_error: bool = ignore_translated_error
        self._max_retries: int = max_retries
        self._max_fill_displaying_errors: int = max_fill_displaying_errors

    def translate_to_element(self, element: Element) -> Element:
        for translated, _, _ in self.translate_to_text_segments(((element, None),)):
            return translated
        raise RuntimeError("Translation failed unexpectedly")

    def translate_to_text_segments(
        self, items: Iterable[tuple[Element, T]]
    ) -> Generator[tuple[Element, list[TextSegment], T], None, None]:
        sync: IterSync[tuple[Element, T]] = IterSync()
        text_segments: list[TextSegment] = []

        for text_segment in self._translate_text_segments(
            elements=(e for e, _ in sync.iter(items)),
        ):
            while True:
                if sync.tail is None:
                    break
                tail_element, _ = sync.tail
                if id(tail_element) == id(text_segment.root):
                    break
                tail_element, payload = sync.take()
                yield tail_element, text_segments, payload
                text_segments = []
            text_segments.append(text_segment)

        while sync.tail is not None:
            tail_element, payload = sync.take()
            yield tail_element, text_segments, payload
            text_segments = []

    def _translate_text_segments(self, elements: Iterable[Element]):
        for group in self._group_context.split_groups(elements):
            text_segments = list(group)
            fill = XMLFill(text_segments)
            source_text = "".join(self._render_text_segments(text_segments))
            translated_text = self._translate_text(source_text)
            self._fill_into_xml(
                fill=fill,
                source_text=source_text,
                translated_text=translated_text,
            )
            yield from group.body

    def _render_text_segments(self, segments: Iterable[TextSegment]):
        iterator = iter(segments)
        segment = next(iterator, None)
        if segment is None:
            return
        while True:
            next_segment = next(iterator, None)
            if next_segment is None:
                break
            yield segment.text
            if id(segment.block_parent) != id(next_segment.block_parent):
                yield "\n\n"
            segment = next_segment
        yield segment.text

    def _translate_text(self, text: str) -> str:
        return self._llm.request(
            input=[
                Message(
                    role=MessageRole.SYSTEM,
                    message=self._llm.template("translate").render(
                        target_language=self._target_language,
                        user_prompt=self._user_prompt,
                    ),
                ),
                Message(role=MessageRole.USER, message=text),
            ]
        )

    def _fill_into_xml(self, fill: XMLFill, source_text: str, translated_text: str) -> Element:
        user_message = (
            f"Source text:\n{source_text}\n\n"
            f"XML template:\n```XML\n{encode_friendly(fill.request_element)}\n```\n\n"
            f"Translated text:\n{translated_text}"
        )
        fixed_messages: list[Message] = [
            Message(
                role=MessageRole.SYSTEM,
                message=self._llm.template("fill").render(),
            ),
            Message(
                role=MessageRole.USER,
                message=user_message,
            ),
        ]

        validator = ProgressiveLockingValidator()
        conversation_history: list[Message] = []
        latest_error: ValidationError | None = None

        for _ in range(self._max_retries):
            # Request LLM response
            response = self._llm.request(
                input=fixed_messages + conversation_history,
            )

            try:
                # Extract XML from response
                validated_element = _extract_xml_element(response)

                # Validate with progressive locking
                is_complete, error_message, newly_locked = validator.validate_with_locking(
                    template_ele=fill.request_element,
                    validated_ele=validated_element,
                    errors_limit=self._max_fill_displaying_errors,
                )

                if is_complete:
                    # All nodes locked, fill successful
                    fill._fill_submitted_texts(  # pylint: disable=protected-access
                        generated_ids_stack=[],
                        element=validated_element,
                    )
                    return validated_element

                # Not complete yet, construct error message with progress info
                progress_msg = f"Progress: {len(validator.locked_ids)} nodes locked"
                if newly_locked:
                    progress_msg += f", {len(newly_locked)} newly locked this round"

                full_error_message = f"{progress_msg}\n\n{error_message}"

                conversation_history = [
                    Message(role=MessageRole.ASSISTANT, message=response),
                    Message(role=MessageRole.USER, message=full_error_message),
                ]

            except ValidationError as error:
                # XML extraction or basic validation failed
                latest_error = error
                conversation_history = [
                    Message(role=MessageRole.ASSISTANT, message=response),
                    Message(role=MessageRole.USER, message=str(error)),
                ]

        message = f"Failed to get valid XML structure after {self._max_retries} attempts"
        if latest_error is None:
            raise ValueError(message)
        else:
            raise ValueError(message) from latest_error



from collections.abc import Callable
from xml.etree.ElementTree import Element

from .math import xml_to_latex

_MATH_TAG = "math"
_EXPRESSION_TAG = "expression"

_PLACEHOLDER_TAGS = frozenset((_EXPRESSION_TAG,))


def is_placeholder_tag(tag: str) -> bool:
    return tag in _PLACEHOLDER_TAGS


class Placeholder:
    def __init__(self, root: Element):
        self._raw_elements: dict[int, Element] = {}
        self._root: Element = self._replace(
            element=root,
            replace=self._replace_raw,
        )
        assert id(self._root) == id(root)

    def recover(self) -> None:
        self._replace(
            element=self._root,
            replace=self._recover_to_raw,
        )

    def _replace(self, element: Element, replace: Callable[[Element], Element | None]) -> Element:
        replaced = replace(element)
        if replaced is not None:
            return replaced
        if len(element):
            element[:] = [self._replace(child, replace) for child in element]
        return element

    def _replace_raw(self, element: Element) -> Element | None:
        if element.tag == _MATH_TAG:
            replaced = Element(_EXPRESSION_TAG)
            replaced.text = xml_to_latex(element)
            replaced.tail = element.tail
            self._raw_elements[id(replaced)] = element
            return replaced
        return None

    def _recover_to_raw(self, replaced: Element) -> Element | None:
        raw = self._raw_elements.get(id(replaced))
        if raw is not None:
            del self._raw_elements[id(replaced)]
            return raw
        return None



CHINESE = "Simplified Chinese"
ENGLISH = "English"
JAPANESE = "Japanese"
KOREAN = "Korean"
SPANISH = "Spanish"
FRENCH = "French"
GERMAN = "German"
PORTUGUESE = "Portuguese"
TRADITIONAL_CHINESE = "Traditional Chinese"
RUSSIAN = "Russian"
ITALIAN = "Italian"
ARABIC = "Arabic"
HINDI = "Hindi"
DUTCH = "Dutch"
POLISH = "Polish"
TURKISH = "Turkish"
VIETNAMESE = "Vietnamese"
THAI = "Thai"
INDONESIAN = "Indonesian"
SWEDISH = "Swedish"
DANISH = "Danish"
NORWEGIAN = "Norwegian"
FINNISH = "Finnish"



from collections.abc import Iterable
from xml.etree.ElementTree import Element

from ..xml import iter_with_stack
from .text_segment import TextPosition, TextSegment, combine_text_segments


def submit_text_segments(element: Element, text_segments: Iterable[TextSegment]):
    grouped_map = _group_text_segments(text_segments)
    flatten_text_segments = dict(_extract_flatten_text_segments(element, grouped_map))
    _append_text_segments(element, grouped_map)
    _replace_text_segments(element, flatten_text_segments)


def _group_text_segments(text_segments: Iterable[TextSegment]):
    grouped_map: dict[int, list[TextSegment]] = {}
    for text_segment in text_segments:
        parent_id = id(text_segment.block_parent)
        grouped = grouped_map.get(parent_id, None)
        if grouped is None:
            grouped_map[parent_id] = grouped = []
        grouped_map[parent_id].append(text_segment)
    return grouped_map


# 被覆盖的 block 表示一种偶然现象，由于它的子元素会触发 append 操作，若对它也进行 append 操作阅读顺序会混乱
# 此时只能在它的所有文本后立即接上翻译后的文本
def _extract_flatten_text_segments(element: Element, grouped_map: dict[int, list[TextSegment]]):
    override_parent_ids: set[int] = set()
    for parents, child_element in iter_with_stack(element):
        if id(child_element) not in grouped_map:
            continue
        for parent in parents[:-1]:
            parent_id = id(parent)
            if parent_id in grouped_map:
                override_parent_ids.add(parent_id)

    if id(element) in grouped_map:
        override_parent_ids.add(id(element))  # root 不会出现在 parents 中需单独添加

    for parent_id in override_parent_ids:
        yield parent_id, grouped_map.pop(parent_id)


def _replace_text_segments(element: Element, text_segments: dict[int, list[TextSegment]]):
    for _, child_element in iter_with_stack(element):
        tail_text_segments: list[TextSegment] = []
        for text_segment in text_segments.get(id(child_element), ()):
            if text_segment.position == TextPosition.TEXT:
                child_element.text = _append_text(
                    origin_text=child_element.text,
                    append_text=text_segment.text,
                )
            elif text_segment.position == TextPosition.TAIL:
                tail_text_segments.append(text_segment)

        tail_text_segments.sort(key=lambda t: t.index)
        tail_text_segments.reverse()
        for cc_element in child_element:
            if not tail_text_segments:
                break
            if cc_element.tail is not None:
                cc_element.tail = _append_text(
                    origin_text=cc_element.tail,
                    append_text=tail_text_segments.pop().text,
                )


def _append_text_segments(element: Element, grouped_map: dict[int, list[TextSegment]]):
    for parents, child_element in iter_with_stack(element):
        if not parents:
            continue
        grouped = grouped_map.get(id(child_element))
        if not grouped:
            continue
        parent = parents[-1]
        index = _index_of_parent(parents[-1], child_element)
        combined = next(
            combine_text_segments(
                segments=(t.strip_block_parents() for t in grouped),
            ),
            None,
        )
        if combined is not None:
            combined_element, _ = combined
            parent.insert(index + 1, combined_element)
            combined_element.tail = child_element.tail
            child_element.tail = None


def _index_of_parent(parent: Element, checked_element: Element) -> int:
    for i, child in enumerate(parent):
        if child == checked_element:
            return i
    raise ValueError("Element not found in parent.")


def _append_text(origin_text: str | None, append_text: str) -> str:
    if origin_text is None:
        return append_text
    else:
        return origin_text + append_text



from collections.abc import Generator, Iterable
from dataclasses import dataclass
from xml.etree.ElementTree import Element

from resource_segmentation import Resource, Segment, split
from tiktoken import Encoding

from .fragmented import group_fragmented_elements
from .text_segment import TextSegment, incision_between, search_text_segments

_BORDER_INCISION = 0
_ELLIPSIS = "..."


@dataclass
class XMLGroup:
    head: list[TextSegment]
    body: list[TextSegment]
    tail: list[TextSegment]

    def __iter__(self) -> Generator[TextSegment, None, None]:
        yield from self.head
        yield from self.body
        yield from self.tail


class XMLGroupContext:
    def __init__(self, encoding: Encoding, max_group_tokens: int) -> None:
        self._encoding: Encoding = encoding
        self._max_group_tokens: int = max_group_tokens

    def split_groups(self, elements: Iterable[Element]) -> Generator[XMLGroup, None, None]:
        for grouped_elements in group_fragmented_elements(
            encoding=self._encoding,
            elements=elements,
            group_max_tokens=self._max_group_tokens,
        ):
            for group in split(
                resources=self._expand_text_segments(grouped_elements),
                max_segment_count=self._max_group_tokens,
                border_incision=_BORDER_INCISION,
            ):
                yield XMLGroup(
                    head=list(
                        self._truncate_text_segments(
                            segments=self._expand_text_segments_with_items(group.head),
                            remain_head=False,
                            remain_count=group.head_remain_count,
                        )
                    ),
                    body=list(self._expand_text_segments_with_items(group.body)),
                    tail=list(
                        self._truncate_text_segments(
                            segments=self._expand_text_segments_with_items(group.tail),
                            remain_head=True,
                            remain_count=group.tail_remain_count,
                        )
                    ),
                )

    def _expand_text_segments(self, elements: Iterable[Element]):
        for element in elements:
            yield from self._expand_text_segments_with_element(element)

    def _expand_text_segments_with_element(self, element: Element) -> Generator[Resource[TextSegment], None, None]:
        generator = search_text_segments(element)
        segment = next(generator, None)
        start_incision = _BORDER_INCISION
        if segment is None:
            return

        while True:
            next_segment = next(generator, None)
            if next_segment is None:
                break
            incision1, incision2 = incision_between(
                segment1=segment,
                segment2=next_segment,
            )
            yield Resource(
                count=len(self._encoding.encode(segment.xml_text)),
                start_incision=start_incision,
                end_incision=incision1,
                payload=segment,
            )
            segment = next_segment
            start_incision = incision2

        yield Resource(
            count=len(self._encoding.encode(segment.xml_text)),
            start_incision=start_incision,
            end_incision=_BORDER_INCISION,
            payload=segment,
        )

    def _expand_text_segments_with_items(self, items: list[Resource[TextSegment] | Segment[TextSegment]]):
        for item in items:
            if isinstance(item, Resource):
                yield item.payload.clone()
            elif isinstance(item, Segment):
                for resource in item.resources:
                    yield resource.payload.clone()

    def _truncate_text_segments(self, segments: Iterable[TextSegment], remain_head: bool, remain_count: int):
        if remain_head:
            yield from self._filter_and_remain_segments(
                segments=segments,
                remain_head=remain_head,
                remain_count=remain_count,
            )
        else:
            yield from reversed(
                list(
                    self._filter_and_remain_segments(
                        segments=reversed(list(segments)),
                        remain_head=remain_head,
                        remain_count=remain_count,
                    )
                )
            )

    def _filter_and_remain_segments(self, segments: Iterable[TextSegment], remain_head: bool, remain_count: int):
        for segment in segments:
            if remain_count <= 0:
                break
            raw_xml_text = segment.xml_text
            tokens = self._encoding.encode(raw_xml_text)
            tokens_count = len(tokens)

            if tokens_count > remain_count:
                truncated_segment = self._truncate_text_segment(
                    segment=segment,
                    tokens=tokens,
                    raw_xml_text=raw_xml_text,
                    remain_head=remain_head,
                    remain_count=remain_count,
                )
                if truncated_segment is not None:
                    yield truncated_segment
                break

            yield segment
            remain_count -= tokens_count

    def _truncate_text_segment(
        self,
        segment: TextSegment,
        tokens: list[int],
        raw_xml_text: str,
        remain_head: bool,
        remain_count: int,
    ) -> TextSegment | None:
        # 典型的 xml_text: <tag id="99" data-origin-len="999">Some text</tag>
        # 如果切割点在前缀 XML 区，则整体舍弃
        # 如果切割点在后缀 XML 区，则整体保留
        # 只有刚好切割在正文区，才执行文本截断操作
        remain_text: str
        xml_text_head_length = raw_xml_text.find(segment.text)

        if remain_head:
            remain_xml_text = self._encoding.decode(tokens[:remain_count])  # remain_count cannot be 0 here
            if len(remain_xml_text) <= xml_text_head_length:
                return
            if len(remain_xml_text) >= xml_text_head_length + len(segment.text):
                return segment
            remain_text = remain_xml_text[xml_text_head_length:]
        else:
            xml_text_tail_length = len(raw_xml_text) - (xml_text_head_length + len(segment.text))
            remain_xml_text = self._encoding.decode(tokens[-remain_count:])
            if len(remain_xml_text) <= xml_text_tail_length:
                return
            if len(remain_xml_text) >= xml_text_tail_length + len(segment.text):
                return segment
            remain_text = remain_xml_text[: len(remain_xml_text) - xml_text_tail_length]

        if not remain_text.strip():
            return

        if remain_head:
            segment.text = f"{remain_text} {_ELLIPSIS}"
        else:
            segment.text = f"{_ELLIPSIS} {remain_text}"
        return segment



from dataclasses import dataclass
from enum import Enum, auto
from typing import TypeVar

R = TypeVar("R")


@dataclass
class Message:
    role: "MessageRole"
    message: str


class MessageRole(Enum):
    SYSTEM = auto()
    USER = auto()
    ASSISTANT = auto()
